"""Constants that specify sample sources"""

METAGENOME_SOURCES = (
    "blood",
    "skin",
    "respiratory",
    "urine",
    "CSF",
    "faeces",
    "environmental",
    "unknown",
    "other",
)
ANALYSIS_SOURCES = (
    "blood",
    "buccal swab",
    "cell-free DNA",
    "cell line",
    "cytology (FFPE)",
    "cytology (not fixed/fresh)",
    "muscle",
    "nail",
    "saliva",
    "skin",
    "tissue (FFPE)",
    "tissue (fresh frozen)",
    "bone marrow",
    "other",
)



from cg.store import Store
from cg.store import models


def test_store_api_delete_flowcell(flow_cell_id: str, populated_flow_cell_store: Store):
    """Test function to delete a flow cell entry in Store"""

    # GIVEN a database containing a flow cell
    flow_cell: models.Flowcell = populated_flow_cell_store.Flowcell.query.filter(
        models.Flowcell.name == flow_cell_id
    ).first()

    assert flow_cell

    # WHEN removing said flow cell
    populated_flow_cell_store.delete_flowcell(flowcell_name=flow_cell_id)

    # THEN no entry should be found for said flow cell
    results: models.Flowcell = populated_flow_cell_store.Flowcell.query.filter(
        models.Flowcell.name == flow_cell_id
    ).first()

    assert not results



from pathlib import Path
from cg.apps.cgstats.parsers.quality_metrics import QualityMetrics


def test_parse_quality_metrics(quality_metrics_path: Path):
    # GIVEN an existing Quality_Metrics.csv path
    assert quality_metrics_path.exists()

    # WHEN instantiation a QualityMetrics obj
    quality_metrics_obj: QualityMetrics = QualityMetrics(quality_metrics_path)

    # THEN the object should be successfully parsed
    assert quality_metrics_obj.parse_metrics_file()



"""Loqusdb related constants."""

from enum import Enum

from cgmodels.cg.constants import Pipeline, StrEnum

from cg.constants.sequencing import SequencingMethod

LOQUSDB_ID = "_id"
LOQUSDB_SUPPORTED_PIPELINES = [Pipeline.MIP_DNA, Pipeline.BALSAMIC]
LOQUSDB_MIP_SEQUENCING_METHODS = [SequencingMethod.WGS, SequencingMethod.WES]
LOQUSDB_BALSAMIC_SEQUENCING_METHODS = [SequencingMethod.WGS]


class LoqusdbInstance(StrEnum):
    """Observations instances."""

    WGS: str = "loqusdb"
    WES: str = "loqusdb-wes"
    SOMATIC: str = "loqusdb-somatic"
    TUMOR: str = "loqusdb-tumor"


class MipDNAObservationsAnalysisTag(StrEnum):
    """Rare disease observations files analysis tags."""

    SNV_VCF: str = "deepvariant"
    SV_VCF: str = "vcf-sv-research"
    PROFILE_GBCF: str = "snv-gbcf"
    FAMILY_PED: str = "pedigree"


class MipDNALoadParameters(Enum):
    """Rare disease Loqusdb load command parameters."""

    PROFILE_THRESHOLD: float = 0.95
    GQ_THRESHOLD: int = 10
    HARD_THRESHOLD: float = 0.95
    SOFT_THRESHOLD: float = 0.90


class BalsamicObservationsAnalysisTag(StrEnum):
    """Cancer observations files analysis tags."""

    SNV_VCF: str = "vcf-snv-clinical"
    SNV_ALL_VCF: str = "vcf-snv-germline-tumor"
    SV_VCF: str = "vcf-sv-clinical"
    PROFILE_VCF: str = "vcf-snv-germline-tumor"


class BalsamicLoadParameters(Enum):
    """Cancer Loqusdb load command parameters."""

    PROFILE_THRESHOLD: float = 0.95
    GQ_THRESHOLD: int = 10
    HARD_THRESHOLD: float = 0.95
    SOFT_THRESHOLD: float = 0.90



"""Module that handles deletion of flow cell run directories and their BCL files from
flow_cell_run_dir/<sequencer>."""
import logging
import shutil
from datetime import datetime, timedelta
from pathlib import Path
from typing import Optional

from housekeeper.store import models as hk_models

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import FlowCellStatus
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.symbols import UNDERSCORE
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.store import Store

FLOW_CELL_DATE_POSITION = 0
FLOW_CELL_IDENTIFIER_POSITION = 3
LOG = logging.getLogger(__name__)
PDC_RETRIEVAL_STATUSES = [FlowCellStatus.PROCESSING, FlowCellStatus.RETRIEVED]


class RunDirFlowCell:
    """Class to check and remove flow cell run directories"""

    def __init__(self, flow_cell_dir: Path, status_db: Store, housekeeper_api: HousekeeperAPI):
        self.flow_cell_dir: Path = flow_cell_dir
        self.status_db: Store = status_db
        self.hk: HousekeeperAPI = housekeeper_api
        self.flow_cell_name: str = self.flow_cell_dir.name
        self.identifier: str = self.flow_cell_name.split(UNDERSCORE)[FLOW_CELL_IDENTIFIER_POSITION]
        self.run_date: str = self.flow_cell_name.split(UNDERSCORE)[FLOW_CELL_DATE_POSITION]
        self.derived_date: datetime = datetime.strptime(self.run_date, "%y%m%d")
        self.id: str = self.identifier[1:]
        self._age: Optional[timedelta] = None
        self._is_retrieved_from_pdc: Optional[bool] = None
        self._exists_in_statusdb: Optional[bool] = None
        self._sequenced_date: Optional[datetime] = None
        self._flow_cell_status: Optional[str] = None
        self.sample_sheet_path: Path = (
            self.flow_cell_dir / DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
        )

    @property
    def age(self) -> int:
        """How long ago a flow cell was sequenced, in days"""
        if self._age is None:
            self._age: int = (datetime.now() - self.sequenced_date).days
        return self._age

    @property
    def sequenced_date(self) -> datetime:
        """The date on which the flow cell was sequenced"""
        if self._sequenced_date is None:
            if self.status_db.get_flow_cell(flow_cell_id=self.id):
                LOG.info("Found flow cell %s in statusdb, getting sequenced date.", self.id)
                self._sequenced_date: datetime = self.status_db.get_flow_cell(
                    flow_cell_id=self.id
                ).sequenced_at
            else:
                LOG.info(
                    "Flow cell %s NOT found in statusdb, deriving sequenced date from run dir name!",
                    self.id,
                )
                self._sequenced_date: datetime = self.derived_date
        return self._sequenced_date

    @property
    def is_retrieved_from_pdc(self) -> bool:
        """Flow cell is (being) retrieved from PDC"""
        if self._is_retrieved_from_pdc is None:
            self._is_retrieved_from_pdc = self.flow_cell_status in PDC_RETRIEVAL_STATUSES
        return self._is_retrieved_from_pdc

    @property
    def flow_cell_status(self) -> str:
        """Status of the flow cell"""
        if self._flow_cell_status is None:
            self._flow_cell_status = self.status_db.get_flow_cell(flow_cell_id=self.id).status
        return self._flow_cell_status

    @property
    def exists_in_statusdb(self) -> bool:
        """The flow cell exists in statusdb"""
        if self._exists_in_statusdb is None:
            self._exists_in_statusdb = (
                self.status_db.get_flow_cell(flow_cell_id=self.id) is not None
            )
        return self._exists_in_statusdb

    def remove_run_directory(self):
        """Removes the flow cell run directory"""
        LOG.info("Removing run directory %s", self.flow_cell_dir)
        shutil.rmtree(self.flow_cell_dir, ignore_errors=True)

    def archive_sample_sheet(self):
        """Archives a sample sheet in housekeeper-bundles"""
        LOG.info("Start archiving sample sheet %s", self.sample_sheet_path)
        if not self.sample_sheet_path.exists():
            LOG.warning("Sample sheet does not exists!")
            return
        LOG.info("Sample sheet found!")
        hk_bundle: hk_models.Bundle = self.hk.bundle(name=self.id)
        if hk_bundle is None:
            LOG.info("Creating bundle with name %s", self.id)
            self.hk.create_new_bundle_and_version(name=self.id)
        try:
            self.hk.add_and_include_file_to_latest_version(
                bundle_name=self.id,
                file=self.sample_sheet_path,
                tags=[SequencingFileTag.ARCHIVED_SAMPLE_SHEET, self.id],
            )
        except FileExistsError:
            LOG.warning("Sample sheet already included!")



from cgmodels.cg.constants import Pipeline

from cg.meta.workflow.mip import MipAnalysisAPI
from cg.models.mip.mip_analysis import MipAnalysis


def test_instantiate_mip_analysis(mip_analysis_raw: dict):
    """
    Tests raw mip analysis against a pydantic MipAnalysis
    """
    # GIVEN a dictionary with the some metrics

    # WHEN instantiating a MipAnalysis object
    mip_dna_analysis_object = MipAnalysis(**mip_analysis_raw)

    # THEN assert that it was successfully created
    assert isinstance(mip_dna_analysis_object, MipAnalysis)


def test_instantiate_parse_mip_analysis(
    cg_context,
    mip_analysis_config_dna_raw: dict,
    mip_metrics_deliverables_raw: dict,
    sample_info_dna_raw: dict,
):
    """
    Tests parse_analysis
    """
    # GIVEN a dictionary with some metrics and a MIP analysis API
    mip_analysis_api = MipAnalysisAPI(cg_context, Pipeline.MIP_DNA)

    # WHEN instantiating a MipAnalysis object
    mip_dna_analysis_object = mip_analysis_api.parse_analysis(
        config_raw=mip_analysis_config_dna_raw,
        qc_metrics_raw=mip_metrics_deliverables_raw,
        sample_info_raw=sample_info_dna_raw,
    )

    # THEN assert that it was successfully created
    assert isinstance(mip_dna_analysis_object, MipAnalysis)



"""Conftest file for pytest fixtures that needs to be shared for multiple tests."""
import copy

import datetime as dt
import logging
import os
import shutil
from pathlib import Path
from typing import Generator, Dict, List, Any

import pytest
from housekeeper.store.models import File

from cg.apps.gt import GenotypeAPI
from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import Pipeline
from cg.constants.constants import FileFormat
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles, BclConverter
from cg.constants.priority import SlurmQos
from cg.io.controller import ReadFile
from cg.constants.subject import Gender
from cg.meta.rsync import RsyncAPI
from cg.meta.transfer.external_data import ExternalDataAPI
from cg.models import CompressionData
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flow_cell import FlowCell
from cg.store import Store

from .mocks.crunchy import MockCrunchyAPI
from .mocks.hk_mock import MockHousekeeperAPI
from .mocks.limsmock import MockLimsAPI
from .mocks.madeline import MockMadelineAPI
from .mocks.osticket import MockOsTicket
from .mocks.process_mock import ProcessMock
from .mocks.scout import MockScoutAPI
from .mocks.tb_mock import MockTB
from .small_helpers import SmallHelpers
from .store_helpers import StoreHelpers

from housekeeper.store import models as hk_models

LOG = logging.getLogger(__name__)


# Timestamp fixture


@pytest.fixture(scope="function", name="old_timestamp")
def fixture_old_timestamp() -> dt.datetime:
    """Return a time stamp in date time format."""
    return dt.datetime(1900, 1, 1)


@pytest.fixture(scope="function", name="timestamp")
def fixture_timestamp() -> dt.datetime:
    """Return a time stamp in date time format."""
    return dt.datetime(2020, 5, 1)


@pytest.fixture(scope="function", name="later_timestamp")
def fixture_later_timestamp() -> dt.datetime:
    """Return a time stamp in date time format."""
    return dt.datetime(2020, 6, 1)


@pytest.fixture(scope="function", name="timestamp_now")
def fixture_timestamp_now() -> dt.datetime:
    """Return a time stamp of today's date in date time format."""
    return dt.datetime.now()


@pytest.fixture(scope="function", name="timestamp_yesterday")
def fixture_timestamp_yesterday(timestamp_now: dt.datetime) -> dt.datetime:
    """Return a time stamp of yesterday's date in date time format."""
    return timestamp_now - dt.timedelta(days=1)


@pytest.fixture(scope="function", name="timestamp_in_2_weeks")
def fixture_timestamp_in_2_weeks(timestamp_now: dt.datetime) -> dt.datetime:
    """Return a time stamp 14 days ahead in time."""
    return timestamp_now + dt.timedelta(days=14)


# Case fixtures


@pytest.fixture(name="slurm_account")
def fixture_slurm_account() -> str:
    """Return a SLURM account."""
    return "super_account"


@pytest.fixture(name="user_name")
def fixture_user_name() -> str:
    """Return a user name."""
    return "Paul Anderson"


@pytest.fixture(name="user_mail")
def fixture_user_mail() -> str:
    """Return a user email."""
    return "paul@magnolia.com"


@pytest.fixture(name="email_adress")
def fixture_email_adress() -> str:
    """Return an email adress."""
    return "james.holden@scilifelab.se"


@pytest.fixture(name="case_id")
def fixture_case_id() -> str:
    """Return a case id."""
    return "yellowhog"


@pytest.fixture(name="another_case_id")
def fixture_another_case_id() -> str:
    """Return another case id."""
    return "another_case_id"


@pytest.fixture(name="sample_id")
def fixture_sample_id() -> str:
    """Returns a sample id."""
    return "ADM1"


@pytest.fixture(name="sample_name")
def fixture_sample_name() -> str:
    """Returns a sample name."""
    return "a_sample_name"


@pytest.fixture(name="cust_sample_id", scope="session")
def fixture_cust_sample_id() -> str:
    """Returns a customer sample id."""
    return "child"


@pytest.fixture(name="family_name")
def fixture_family_name() -> str:
    """Return a case name."""
    return "case"


@pytest.fixture(name="customer_id", scope="session")
def fixture_customer_id() -> str:
    """Return a customer id."""
    return "cust000"


@pytest.fixture(name="sbatch_job_number")
def fixture_sbatch_job_number() -> int:
    return 123456


@pytest.fixture(name="sbatch_process")
def fixture_sbatch_process(sbatch_job_number: int) -> ProcessMock:
    """Return a mocked process object."""
    slurm_process = ProcessMock(binary="sbatch")
    slurm_process.set_stdout(text=str(sbatch_job_number))
    return slurm_process


@pytest.fixture(scope="function", name="analysis_family_single_case")
def fixture_analysis_family_single(
    case_id: str, family_name: str, sample_id: str, ticket: str
) -> dict:
    """Build an example case."""
    return {
        "name": family_name,
        "internal_id": case_id,
        "data_analysis": str(Pipeline.MIP_DNA),
        "application_type": "wgs",
        "panels": ["IEM", "EP"],
        "tickets": ticket,
        "samples": [
            {
                "name": "proband",
                "sex": Gender.MALE,
                "internal_id": sample_id,
                "status": "affected",
                "original_ticket": ticket,
                "reads": 5000000000,
                "capture_kit": "GMSmyeloid",
            }
        ],
    }


@pytest.fixture(scope="function", name="analysis_family")
def fixture_analysis_family(case_id: str, family_name: str, sample_id: str, ticket: str) -> dict:
    """Return a dictionary with information from a analysis case."""
    return {
        "name": family_name,
        "internal_id": case_id,
        "data_analysis": str(Pipeline.MIP_DNA),
        "application_type": "wgs",
        "tickets": ticket,
        "panels": ["IEM", "EP"],
        "samples": [
            {
                "name": "child",
                "sex": Gender.MALE,
                "internal_id": sample_id,
                "father": "ADM2",
                "mother": "ADM3",
                "status": "affected",
                "original_ticket": ticket,
                "reads": 5000000,
                "capture_kit": "GMSmyeloid",
            },
            {
                "name": "father",
                "sex": Gender.MALE,
                "internal_id": "ADM2",
                "status": "unaffected",
                "original_ticket": ticket,
                "reads": 6000000,
                "capture_kit": "GMSmyeloid",
            },
            {
                "name": "mother",
                "sex": Gender.FEMALE,
                "internal_id": "ADM3",
                "status": "unaffected",
                "original_ticket": ticket,
                "reads": 7000000,
                "capture_kit": "GMSmyeloid",
            },
        ],
    }


# Config fixtures


@pytest.fixture(name="base_config_dict")
def fixture_base_config_dict() -> dict:
    """Returns the basic configs necessary for running CG."""
    return {
        "database": "sqlite:///",
        "madeline_exe": "path/to/madeline",
        "bed_path": "path/to/bed",
        "pon_path": "path/to/pon",
        "delivery_path": "path/to/delivery",
        "housekeeper": {
            "database": "sqlite:///",
            "root": "path/to/root",
        },
        "email_base_settings": {
            "sll_port": 465,
            "smtp_server": "smtp.gmail.com",
            "sender_email": "test@gmail.com",
            "sender_password": "",
        },
        "loqusdb": {
            "binary_path": "binary",
            "config_path": "config",
        },
        "loqusdb-wes": {
            "binary_path": "binary_wes",
            "config_path": "config_wes",
        },
        "loqusdb-somatic": {
            "binary_path": "binary_somatic",
            "config_path": "config_somatic",
        },
        "loqusdb-tumor": {
            "binary_path": "binary_tumor",
            "config_path": "config_tumor",
        },
    }


@pytest.fixture(name="cg_config_object")
def fixture_cg_config_object(base_config_dict: dict) -> CGConfig:
    """Return a CG config dict."""
    return CGConfig(**base_config_dict)


@pytest.fixture(name="chanjo_config")
def fixture_chanjo_config() -> Dict[str, Dict[str, str]]:
    """Chanjo configs"""
    return {"chanjo": {"config_path": "chanjo_config", "binary_path": "chanjo"}}


@pytest.fixture
def crunchy_config_dict():
    """Crunchy configs."""
    return {
        "crunchy": {
            "conda_binary": "a conda binary",
            "cram_reference": "/path/to/fasta",
            "slurm": {"account": "mock_account", "mail_user": "mock_mail", "conda_env": "mock_env"},
        }
    }


@pytest.fixture(name="hk_config_dict")
def fixture_hk_config_dict(root_path):
    """Crunchy configs."""
    return {
        "housekeeper": {
            "database": "sqlite:///:memory:",
            "root": str(root_path),
        }
    }


@pytest.fixture(name="genotype_config")
def fixture_genotype_config() -> dict:
    """Genotype config fixture."""
    return {
        "genotype": {
            "database": "database",
            "config_path": "config/path",
            "binary_path": "gtdb",
        }
    }


# Api fixtures


@pytest.fixture(name="rsync_api")
def fixture_rsync_api(cg_context: CGConfig) -> RsyncAPI:
    """RsyncAPI fixture."""
    return RsyncAPI(config=cg_context)


@pytest.fixture(name="external_data_api")
def fixture_external_data_api(analysis_store, cg_context: CGConfig) -> ExternalDataAPI:
    """ExternalDataAPI fixture."""
    return ExternalDataAPI(config=cg_context)


@pytest.fixture(name="genotype_api")
def fixture_genotype_api(genotype_config: dict) -> GenotypeAPI:
    """Genotype API fixture."""
    _genotype_api = GenotypeAPI(genotype_config)
    _genotype_api.set_dry_run(True)
    return _genotype_api


@pytest.fixture(scope="function")
def madeline_api(madeline_output) -> MockMadelineAPI:
    """madeline_api fixture."""
    _api = MockMadelineAPI()
    _api.set_outpath(madeline_output)
    return _api


@pytest.fixture(name="ticket", scope="session")
def fixture_ticket_number() -> str:
    """Return a ticket number for testing."""
    return "123456"


@pytest.fixture(name="osticket")
def fixture_os_ticket(ticket: str) -> MockOsTicket:
    """Return a api that mock the os ticket api."""
    api = MockOsTicket()
    api.set_ticket_nr(ticket)
    return api


# Files fixtures

# Common file fixtures
@pytest.fixture(scope="session", name="fixtures_dir")
def fixture_fixtures_dir() -> Path:
    """Return the path to the fixtures dir."""
    return Path("tests", "fixtures")


@pytest.fixture(name="analysis_dir")
def fixture_analysis_dir(fixtures_dir: Path) -> Path:
    """Return the path to the analysis dir."""
    return Path(fixtures_dir, "analysis")


@pytest.fixture(name="apps_dir")
def fixture_apps_dir(fixtures_dir: Path) -> Path:
    """Return the path to the apps dir."""
    return Path(fixtures_dir, "apps")


@pytest.fixture(name="cgweb_orders_dir", scope="session")
def fixture_cgweb_orders_dir(fixtures_dir: Path) -> Path:
    """Return the path to the cgweb_orders dir."""
    return Path(fixtures_dir, "cgweb_orders")


@pytest.fixture(name="fastq_dir")
def fixture_fastq_dir(demultiplexed_runs: Path) -> Path:
    """Return the path to the fastq files dir."""
    return Path(demultiplexed_runs, "fastq")


@pytest.fixture(scope="function", name="project_dir")
def fixture_project_dir(tmpdir_factory) -> Generator[Path, None, None]:
    """Path to a temporary directory where intermediate files can be stored."""
    yield Path(tmpdir_factory.mktemp("data"))


@pytest.fixture(scope="function")
def tmp_file(project_dir) -> Path:
    """Return a temp file path."""
    return Path(project_dir, "test")


@pytest.fixture(name="non_existing_file_path")
def fixture_non_existing_file_path(project_dir: Path) -> Path:
    """Return the path to a non existing file."""
    return Path(project_dir, "a_file.txt")


@pytest.fixture(name="content")
def fixture_content() -> str:
    """Return some content for a file."""
    return (
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt"
        " ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ull"
        "amco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehende"
        "rit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaec"
        "at cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."
    )


@pytest.fixture(name="filled_file")
def fixture_filled_file(non_existing_file_path: Path, content: str) -> Path:
    """Return the path to a existing file with some content."""
    with open(non_existing_file_path, "w") as outfile:
        outfile.write(content)
    return non_existing_file_path


@pytest.fixture(scope="session", name="orderforms")
def fixture_orderform(fixtures_dir: Path) -> Path:
    """Return the path to the directory with order forms."""
    return Path(fixtures_dir, "orderforms")


@pytest.fixture(name="hk_file")
def fixture_hk_file(filled_file, case_id) -> File:
    """Return a housekeeper File object."""
    return File(id=case_id, path=filled_file)


@pytest.fixture(name="mip_dna_store_files")
def fixture_mip_dna_store_files(apps_dir: Path) -> Path:
    """Return the path to the directory with mip dna store files."""
    return Path(apps_dir, "mip", "dna", "store")


@pytest.fixture(name="case_qc_sample_info_path")
def fixture_case_qc_sample_info_path(mip_dna_store_files: Path) -> Path:
    """Return path to case_qc_sample_info.yaml."""
    return Path(mip_dna_store_files, "case_qc_sample_info.yaml")


@pytest.fixture(name="delivery_report_html")
def fixture_delivery_report_html(mip_dna_store_files: Path) -> Path:
    """Return the path to a qc metrics deliverables file with case data."""
    return Path(mip_dna_store_files, "empty_delivery_report.html")


@pytest.fixture(name="mip_deliverables_file")
def fixture_mip_deliverables_files(mip_dna_store_files: Path) -> Path:
    """Fixture for general deliverables file in mip."""
    return Path(mip_dna_store_files, "case_id_deliverables.yaml")


@pytest.fixture(name="case_qc_metrics_deliverables")
def fixture_case_qc_metrics_deliverables(apps_dir: Path) -> Path:
    """Return the path to a qc metrics deliverables file with case data."""
    return Path(apps_dir, "mip", "case_metrics_deliverables.yaml")


@pytest.fixture(name="mip_analysis_dir")
def fixture_mip_analysis_dir(analysis_dir: Path) -> Path:
    """Return the path to the directory with mip analysis files."""
    return Path(analysis_dir, "mip")


@pytest.fixture(name="balsamic_analysis_dir")
def fixture_balsamic_analysis_dir(analysis_dir: Path) -> Path:
    """Return the path to the directory with balsamic analysis files."""
    return Path(analysis_dir, "balsamic")


@pytest.fixture(name="balsamic_wgs_analysis_dir")
def fixture_balsamic_wgs_analysis_dir(balsamic_analysis_dir: Path) -> Path:
    """Return the path to the directory with balsamic analysis files."""
    return Path(balsamic_analysis_dir, "tn_wgs")


@pytest.fixture(name="mip_dna_analysis_dir")
def fixture_mip_dna_analysis_dir(mip_analysis_dir: Path) -> Path:
    """Return the path to the directory with mip dna analysis files."""
    return Path(mip_analysis_dir, "dna")


@pytest.fixture(name="sample1_cram")
def fixture_sample1_cram(mip_dna_analysis_dir: Path) -> Path:
    """Return the path to the cram file for sample 1."""
    return Path(mip_dna_analysis_dir, "adm1.cram")


@pytest.fixture(name="vcf_file")
def fixture_vcf_file(mip_dna_store_files: Path) -> Path:
    """Return the path to to a vcf file."""
    return Path(mip_dna_store_files, "yellowhog_clinical_selected.vcf")


@pytest.fixture(name="fastq_file")
def fixture_fastq_file(fastq_dir: Path) -> Path:
    """Return the path to to a fastq file."""
    return Path(fastq_dir, "dummy_run_R1_001.fastq.gz")


@pytest.fixture(name="madeline_output")
def fixture_madeline_output(apps_dir: Path) -> Path:
    """Return str of path for file with Madeline output."""
    return Path(apps_dir, "madeline", "madeline.xml")


@pytest.fixture(name="file_does_not_exist")
def fixture_file_does_not_exist() -> Path:
    """Return a file path that does not exist."""
    return Path("file", "does", "not", "exist")


# Compression fixtures


@pytest.fixture(scope="function", name="run_name")
def fixture_run_name() -> str:
    """Return the name of a fastq run."""
    return "fastq_run"


@pytest.fixture(scope="function", name="original_fastq_data")
def fixture_original_fastq_data(fastq_dir: Path, run_name) -> CompressionData:
    """Return a compression object with a path to the original fastq files."""
    return CompressionData(Path(fastq_dir, run_name))


@pytest.fixture(scope="function", name="fastq_stub")
def fixture_fastq_stub(project_dir: Path, run_name: str) -> Path:
    """Creates a path to the base format of a fastq run."""
    return Path(project_dir, run_name)


@pytest.fixture(scope="function", name="compression_object")
def fixture_compression_object(
    fastq_stub: Path, original_fastq_data: CompressionData
) -> CompressionData:
    """Creates compression data object with information about files used in fastq compression."""
    working_files: CompressionData = CompressionData(fastq_stub)
    working_file_map: Dict[str, str] = {
        original_fastq_data.fastq_first.as_posix(): working_files.fastq_first.as_posix(),
        original_fastq_data.fastq_second.as_posix(): working_files.fastq_second.as_posix(),
    }
    for original_file, working_file in working_file_map.items():
        shutil.copy(original_file, working_file)
    return working_files


# Demultiplex fixtures


@pytest.fixture(name="demultiplex_fixtures")
def fixture_demultiplex_fixtures(apps_dir: Path) -> Path:
    """Return the path to the demultiplex fixtures."""
    return Path(apps_dir, "demultiplexing")


@pytest.fixture(name="novaseq_dragen_sample_sheet_path")
def fixture_novaseq_dragen_sample_sheet_path(demultiplex_fixtures: Path) -> Path:
    """Return the path to a novaseq bcl2fastq sample sheet."""
    return Path(demultiplex_fixtures, "SampleSheetS2_Dragen.csv")


@pytest.fixture(name="raw_lims_sample_dir")
def fixture_raw_lims_sample_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the raw samples fixtures."""
    return Path(demultiplex_fixtures, "raw_lims_samples")


@pytest.fixture(name="demultiplexed_runs")
def fixture_demultiplexed_runs(demultiplex_fixtures: Path) -> Path:
    """Return the path to a dir with flow cells ready for demultiplexing."""
    return Path(demultiplex_fixtures, "demultiplexed-runs")


@pytest.fixture(name="demux_run_dir")
def fixture_demux_run_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to a dir with flow cells ready for demultiplexing."""
    return Path(demultiplex_fixtures, "flowcell-runs")


@pytest.fixture(name="flow_cell")
def fixture_flow_cell(demux_run_dir: Path, flow_cell_full_name: str) -> FlowCell:
    """Create a flow cell object with flow cell that is demultiplexed."""
    return FlowCell(flow_cell_path=Path(demux_run_dir, flow_cell_full_name))


@pytest.fixture(name="flow_cell_id")
def fixture_flow_cell_id(flow_cell: FlowCell) -> str:
    """Return flow cell id from flow cell object."""
    return flow_cell.id


@pytest.fixture(name="another_flow_cell_id")
def fixture_another_flow_cell_id() -> str:
    """Return another flow cell id."""
    return "HF57HDRXY"


@pytest.fixture(name="demultiplexing_delivery_file")
def fixture_demultiplexing_delivery_file(flow_cell: FlowCell) -> Path:
    """Return demultiplexing delivery started file."""
    return Path(flow_cell.path, DemultiplexingDirsAndFiles.DELIVERY)


@pytest.fixture(name="hiseq_x_tile_dir")
def fixture_hiseq_x_tile_dir(flow_cell: FlowCell) -> Path:
    """Return Hiseq X tile dir."""
    return Path(flow_cell.path, DemultiplexingDirsAndFiles.HiseqX_TILE_DIR)


@pytest.fixture(name="lims_novaseq_samples_file")
def fixture_lims_novaseq_samples_file(raw_lims_sample_dir: Path) -> Path:
    """Return the path to a file with sample info in lims format."""
    return Path(raw_lims_sample_dir, "raw_samplesheet_novaseq.json")


@pytest.fixture(name="lims_novaseq_samples_raw")
def fixture_lims_novaseq_samples_raw(lims_novaseq_samples_file: Path) -> List[dict]:
    """Return a list of raw flow cell samples."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=lims_novaseq_samples_file
    )


@pytest.fixture(name="flow_cell_full_name")
def fixture_flow_cell_full_name() -> str:
    """Return full flow cell name."""
    return "201203_A00689_0200_AHVKJCDRXX"


@pytest.fixture(name="demultiplexed_flow_cell")
def fixture_demultiplexed_flow_cell(demultiplexed_runs: Path, flow_cell_full_name: str) -> Path:
    return Path(demultiplexed_runs, flow_cell_full_name)


@pytest.fixture(name="bcl2fastq_demux_results")
def fixture_bcl2fastq_demux_results(
    demultiplexed_flow_cell: Path, flow_cell: FlowCell
) -> DemuxResults:
    return DemuxResults(
        demux_dir=demultiplexed_flow_cell, flow_cell=flow_cell, bcl_converter=BclConverter.BCL2FASTQ
    )


# Genotype file fixture


@pytest.fixture(name="bcf_file")
def fixture_bcf_file(apps_dir: Path) -> Path:
    """Return the path to a BCF file."""
    return Path(apps_dir, "gt", "yellowhog.bcf")


# Housekeeper, Chanjo file fixtures


@pytest.fixture(scope="function", name="bed_file")
def fixture_bed_file(analysis_dir) -> Path:
    """Return the path to a bed file."""
    return Path(analysis_dir, "sample_coverage.bed")


# Helper fixtures


@pytest.fixture(name="helpers")
def fixture_helpers() -> StoreHelpers:
    """Return a class with helper functions for the stores."""
    return StoreHelpers()


@pytest.fixture(name="small_helpers")
def fixture_small_helpers() -> SmallHelpers:
    """Return a class with small helper functions."""
    return SmallHelpers()


# HK fixtures


@pytest.fixture(name="root_path")
def fixture_root_path(project_dir: Path) -> Path:
    """Return the path to a hk bundles dir."""
    _root_path = project_dir / "bundles"
    _root_path.mkdir(parents=True, exist_ok=True)
    return _root_path


@pytest.fixture(scope="function", name="hk_bundle_data")
def fixture_hk_bundle_data(case_id: str, bed_file: Path, timestamp: dt.datetime) -> Dict[str, Any]:
    """Return some bundle data for Housekeeper."""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [{"path": bed_file.as_posix(), "archive": False, "tags": ["bed", "sample"]}],
    }


@pytest.fixture(scope="function", name="sample_hk_bundle_no_files")
def fixture_sample_hk_bundle_no_files(sample_id: str, timestamp: dt.datetime) -> dict:
    """Create a complete bundle mock for testing compression."""
    return {
        "name": sample_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [],
    }


@pytest.fixture(scope="function", name="case_hk_bundle_no_files")
def fixture_case_hk_bundle_no_files(case_id: str, timestamp: dt.datetime) -> dict:
    """Create a complete bundle mock for testing compression."""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [],
    }


@pytest.fixture(scope="function", name="compress_hk_fastq_bundle")
def fixture_compress_hk_fastq_bundle(
    compression_object: CompressionData, sample_hk_bundle_no_files: dict
) -> dict:
    """Create a complete bundle mock for testing compression

    This bundle contains a pair of fastq files.
    ."""
    hk_bundle_data = copy.deepcopy(sample_hk_bundle_no_files)

    first_fastq = compression_object.fastq_first
    second_fastq = compression_object.fastq_second
    for fastq_file in [first_fastq, second_fastq]:
        fastq_file.touch()
        # We need to set the time to an old date
        # Create a older date
        # Convert the date to a float
        before_timestamp = dt.datetime.timestamp(dt.datetime(2020, 1, 1))
        # Update the utime so file looks old
        os.utime(fastq_file, (before_timestamp, before_timestamp))
        fastq_file_info = {"path": str(fastq_file), "archive": False, "tags": ["fastq"]}

        hk_bundle_data["files"].append(fastq_file_info)
    return hk_bundle_data


@pytest.fixture(name="housekeeper_api")
def fixture_housekeeper_api(hk_config_dict: dict) -> MockHousekeeperAPI:
    """Setup Housekeeper store."""
    return MockHousekeeperAPI(hk_config_dict)


@pytest.fixture(scope="function", name="real_housekeeper_api")
def fixture_real_housekeeper_api(hk_config_dict: dict) -> HousekeeperAPI:
    """Setup a real Housekeeper store."""
    _api = HousekeeperAPI(hk_config_dict)
    _api.initialise_db()
    yield _api


@pytest.fixture(scope="function", name="populated_housekeeper_api")
def fixture_populated_housekeeper_api(
    housekeeper_api: MockHousekeeperAPI, hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Setup a Housekeeper store with some data."""
    hk_api = housekeeper_api
    helpers.ensure_hk_bundle(hk_api, hk_bundle_data)
    return hk_api


@pytest.fixture(scope="function", name="hk_version_obj")
def fixture_hk_version_obj(
    housekeeper_api: MockHousekeeperAPI, hk_bundle_data: dict, helpers
) -> hk_models.Version:
    """Get a Housekeeper version object."""
    return helpers.ensure_hk_version(housekeeper_api, hk_bundle_data)


# Process Mock


@pytest.fixture(name="process")
def fixture_process() -> ProcessMock:
    """Returns a mocked process."""
    return ProcessMock()


# Hermes mock


@pytest.fixture(name="hermes_process")
def fixture_hermes_process() -> ProcessMock:
    """Return a mocked Hermes process."""
    return ProcessMock(binary="hermes")


@pytest.fixture(name="hermes_api")
def fixture_hermes_api(hermes_process: ProcessMock) -> HermesApi:
    """Return a Hermes API with a mocked process."""
    hermes_config = {"hermes": {"binary_path": "/bin/true"}}
    hermes_api = HermesApi(config=hermes_config)
    hermes_api.process = hermes_process
    return hermes_api


# Scout fixtures


@pytest.fixture(scope="function", name="scout_api")
def fixture_scout_api() -> MockScoutAPI:
    """Setup Scout API."""
    return MockScoutAPI()


# Crunchy fixtures


@pytest.fixture(scope="function", name="crunchy_api")
def fixture_crunchy_api():
    """Setup Crunchy API."""
    return MockCrunchyAPI()


# Store fixtures


@pytest.fixture(scope="function", name="analysis_store")
def fixture_analysis_store(
    base_store: Store, analysis_family: dict, wgs_application_tag: str, helpers: StoreHelpers
) -> Generator[Store, None, None]:
    """Setup a store instance for testing analysis API."""
    helpers.ensure_case_from_dict(
        base_store, case_info=analysis_family, app_tag=wgs_application_tag
    )
    yield base_store


@pytest.fixture(scope="function", name="analysis_store_trio")
def fixture_analysis_store_trio(analysis_store: Store) -> Generator[Store, None, None]:
    """Setup a store instance with a trio loaded for testing analysis API."""
    yield analysis_store


@pytest.fixture(scope="function", name="analysis_store_single_case")
def fixture_analysis_store_single(
    base_store: Store, analysis_family_single_case: Store, helpers: StoreHelpers
):
    """Setup a store instance with a single ind case for testing analysis API."""
    helpers.ensure_case_from_dict(base_store, case_info=analysis_family_single_case)
    yield base_store


@pytest.fixture(scope="function", name="collaboration_id")
def fixture_collaboration_id() -> str:
    """Return a default customer group."""
    return "all_customers"


@pytest.fixture(scope="function", name="customer_production")
def fixture_customer_production(collaboration_id: str, customer_id: str) -> dict:
    """Return a dictionary with information about the prod customer."""
    return {
        "customer_id": customer_id,
        "name": "Production",
        "scout_access": True,
        "collaboration_id": collaboration_id,
    }


@pytest.fixture(scope="function", name="external_wgs_application_tag")
def fixture_external_wgs_application_tag() -> str:
    """Return the external WGS application tag."""
    return "WGXCUSC000"


@pytest.fixture(scope="function", name="external_wgs_info")
def fixture_external_wgs_info(external_wgs_application_tag: str) -> dict:
    """Return a dictionary with information external WGS application."""
    return {
        "application_tag": external_wgs_application_tag,
        "application_type": "wgs",
        "description": "External WGS",
        "is_external": True,
        "target_reads": 10,
    }


@pytest.fixture(scope="function", name="external_wes_application_tag")
def fixture_external_wes_application_tag() -> str:
    """Return the external whole exome sequencing application tag."""
    return "EXXCUSR000"


@pytest.fixture(scope="function", name="external_wes_info")
def fixture_external_wes_info(external_wes_application_tag: str) -> dict:
    """Return a dictionary with information external WES application."""
    return {
        "application_tag": external_wes_application_tag,
        "application_type": "wes",
        "description": "External WES",
        "is_external": True,
        "target_reads": 10,
    }


@pytest.fixture(scope="function", name="wgs_application_tag")
def fixture_wgs_application_tag() -> str:
    """Return the WGS application tag."""
    return "WGSPCFC030"


@pytest.fixture(scope="function", name="wgs_application_info")
def fixture_wgs_application_info(wgs_application_tag: str) -> dict:
    """Return a dictionary with information the WGS application."""
    return {
        "application_tag": wgs_application_tag,
        "application_type": "wgs",
        "description": "WGS, double",
        "sequencing_depth": 30,
        "is_external": True,
        "is_accredited": True,
        "target_reads": 10,
    }


@pytest.fixture(name="store")
def fixture_store() -> Store:
    """Fixture with a CG store."""
    _store = Store(uri="sqlite:///")
    _store.create_all()
    yield _store
    _store.drop_all()


@pytest.fixture(name="apptag_rna")
def fixture_apptag_rna() -> str:
    """Return the RNA application tag."""
    return "RNAPOAR025"


@pytest.fixture(scope="function", name="base_store")
def fixture_base_store(store: Store, apptag_rna: str, customer_id: str) -> Store:
    """Setup and example store."""
    collaboration = store.add_collaboration("all_customers", "all customers")

    store.add_commit(collaboration)
    customers = [
        store.add_customer(
            customer_id,
            "Production",
            scout_access=True,
            invoice_address="Test street",
            invoice_reference="ABCDEF",
        ),
        store.add_customer(
            "cust001",
            "Customer",
            scout_access=False,
            invoice_address="Test street",
            invoice_reference="ABCDEF",
        ),
        store.add_customer(
            "cust002",
            "Karolinska",
            scout_access=True,
            invoice_address="Test street",
            invoice_reference="ABCDEF",
        ),
        store.add_customer(
            "cust003",
            "CMMS",
            scout_access=True,
            invoice_address="Test street",
            invoice_reference="ABCDEF",
        ),
    ]
    for customer in customers:
        collaboration.customers.append(customer)
    store.add_commit(customers)
    applications = [
        store.add_application(
            tag="WGXCUSC000",
            category="wgs",
            description="External WGS",
            sequencing_depth=0,
            is_external=True,
            percent_kth=80,
            percent_reads_guaranteed=75,
            target_reads=10,
        ),
        store.add_application(
            tag="EXXCUSR000",
            category="wes",
            description="External WES",
            sequencing_depth=0,
            is_external=True,
            percent_kth=80,
            percent_reads_guaranteed=75,
            target_reads=10,
        ),
        store.add_application(
            tag="WGSPCFC060",
            category="wgs",
            description="WGS, double",
            sequencing_depth=30,
            accredited=True,
            percent_kth=80,
            percent_reads_guaranteed=75,
            target_reads=10,
        ),
        store.add_application(
            tag="RMLP05R800",
            category="rml",
            description="Ready-made",
            sequencing_depth=0,
            percent_kth=80,
            percent_reads_guaranteed=75,
            target_reads=10,
        ),
        store.add_application(
            tag="WGSPCFC030",
            category="wgs",
            description="WGS trio",
            is_accredited=True,
            sequencing_depth=30,
            target_reads=30,
            limitations="some",
            percent_kth=80,
            percent_reads_guaranteed=75,
            min_sequencing_depth=30,
        ),
        store.add_application(
            tag="METLIFR020",
            category="wgs",
            description="Whole genome metagenomics",
            sequencing_depth=0,
            target_reads=400000,
            percent_kth=80,
            percent_reads_guaranteed=75,
        ),
        store.add_application(
            tag="METNXTR020",
            category="wgs",
            description="Metagenomics",
            sequencing_depth=0,
            target_reads=200000,
            percent_kth=80,
            percent_reads_guaranteed=75,
        ),
        store.add_application(
            tag="MWRNXTR003",
            category="mic",
            description="Microbial whole genome ",
            sequencing_depth=0,
            percent_kth=80,
            percent_reads_guaranteed=75,
            target_reads=10,
        ),
        store.add_application(
            tag=apptag_rna,
            category="tgs",
            description="RNA seq, poly-A based priming",
            percent_kth=80,
            percent_reads_guaranteed=75,
            sequencing_depth=25,
            accredited=True,
            target_reads=10,
            min_sequencing_depth=30,
        ),
        store.add_application(
            tag="VWGDPTR001",
            category="cov",
            description="Viral whole genome  ",
            sequencing_depth=0,
            percent_kth=80,
            percent_reads_guaranteed=75,
            target_reads=10,
        ),
    ]

    store.add_commit(applications)

    prices = {"standard": 10, "priority": 20, "express": 30, "research": 5}
    versions = [
        store.add_version(application, 1, valid_from=dt.datetime.now(), prices=prices)
        for application in applications
    ]
    store.add_commit(versions)

    beds = [store.add_bed("Bed")]
    store.add_commit(beds)
    bed_versions = [store.add_bed_version(bed, 1, "Bed.bed") for bed in beds]
    store.add_commit(bed_versions)

    organism = store.add_organism("C. jejuni", "C. jejuni")
    store.add_commit(organism)

    yield store


@pytest.fixture(scope="function")
def sample_store(base_store: Store) -> Store:
    """Populate store with samples."""
    new_samples = [
        base_store.add_sample("ordered", sex=Gender.MALE),
        base_store.add_sample("received", sex=Gender.UNKNOWN, received=dt.datetime.now()),
        base_store.add_sample(
            "received-prepared",
            sex=Gender.UNKNOWN,
            received=dt.datetime.now(),
            prepared_at=dt.datetime.now(),
        ),
        base_store.add_sample("external", sex=Gender.FEMALE, external=True),
        base_store.add_sample(
            "external-received", sex=Gender.FEMALE, received=dt.datetime.now(), external=True
        ),
        base_store.add_sample(
            "sequenced",
            sex=Gender.MALE,
            received=dt.datetime.now(),
            prepared_at=dt.datetime.now(),
            sequenced_at=dt.datetime.now(),
            reads=(310 * 1000000),
        ),
        base_store.add_sample(
            "sequenced-partly",
            sex=Gender.MALE,
            received=dt.datetime.now(),
            prepared_at=dt.datetime.now(),
            reads=(250 * 1000000),
        ),
    ]
    customer = base_store.customers().first()
    external_app = base_store.application("WGXCUSC000").versions[0]
    wgs_app = base_store.application("WGSPCFC030").versions[0]
    for sample in new_samples:
        sample.customer = customer
        sample.application_version = external_app if "external" in sample.name else wgs_app
    base_store.add_commit(new_samples)
    return base_store


@pytest.fixture(scope="function", name="trailblazer_api")
def fixture_trailblazer_api() -> MockTB:
    """Return a mock traailblazer API."""
    return MockTB()


@pytest.fixture(scope="function", name="lims_api")
def fixture_lims_api() -> MockLimsAPI:
    """Return a mock LIMS API."""
    return MockLimsAPI()


@pytest.fixture(name="config_root_dir")
def config_root_dir(tmpdir_factory) -> Path:
    """Return a path to the config root directory."""
    return Path("tests/fixtures/data")


@pytest.fixture()
def housekeeper_dir(tmpdir_factory):
    """Return a temporary directory for Housekeeper testing."""
    return tmpdir_factory.mktemp("housekeeper")


@pytest.fixture()
def mip_dir(tmpdir_factory) -> Path:
    """Return a temporary directory for MIP testing."""
    return tmpdir_factory.mktemp("mip")


@pytest.fixture(scope="function")
def fluffy_dir(tmpdir_factory) -> Path:
    """Return a temporary directory for Fluffy testing."""
    return tmpdir_factory.mktemp("fluffy")


@pytest.fixture(scope="function")
def balsamic_dir(tmpdir_factory) -> Path:
    """Return a temporary directory for Balsamic testing."""
    return tmpdir_factory.mktemp("balsamic")


@pytest.fixture(scope="function")
def cg_dir(tmpdir_factory) -> Path:
    """Return a temporary directory for cg testing."""
    return tmpdir_factory.mktemp("cg")


@pytest.fixture(scope="function")
def microsalt_dir(tmpdir_factory) -> Path:
    """Return a temporary directory for Microsalt testing."""
    return tmpdir_factory.mktemp("microsalt")


@pytest.fixture(name="cg_uri")
def fixture_cg_uri() -> str:
    """Return a cg URI."""
    return "sqlite:///"


@pytest.fixture(name="hk_uri")
def fixture_hk_uri() -> str:
    """Return a Housekeeper URI."""
    return "sqlite:///"


@pytest.fixture(name="loqusdb_id")
def fixture_loqusdb_id() -> str:
    """Returns a Loqusdb mock ID."""
    return "01ab23cd"


@pytest.fixture(name="context_config")
def fixture_context_config(
    cg_uri: str,
    hk_uri: str,
    fluffy_dir: Path,
    housekeeper_dir: Path,
    mip_dir: Path,
    cg_dir: Path,
    balsamic_dir: Path,
    microsalt_dir: Path,
) -> dict:
    """Return a context config."""
    return {
        "bed_path": str(cg_dir),
        "database": cg_uri,
        "delivery_path": str(cg_dir),
        "email_base_settings": {
            "sll_port": 465,
            "smtp_server": "smtp.gmail.com",
            "sender_email": "test@gmail.com",
            "sender_password": "",
        },
        "madeline_exe": "echo",
        "pon_path": str(cg_dir),
        "backup": {
            "encrypt_dir": "/home/ENCRYPT/",
            "root": {"hiseqx": "flowcells/hiseqx", "hiseqga": "RUNS/", "novaseq": "runs/"},
        },
        "balsamic": {
            "balsamic_cache": "hello",
            "binary_path": "echo",
            "conda_env": "S_BALSAMIC",
            "root": str(balsamic_dir),
            "slurm": {
                "account": "development",
                "mail_user": "test.email@scilifelab.se",
                "qos": SlurmQos.LOW,
            },
        },
        "cgstats": {"binary_path": "echo", "database": "sqlite:///./cgstats", "root": str(cg_dir)},
        "chanjo": {"binary_path": "echo", "config_path": "chanjo-stage.yaml"},
        "crunchy": {
            "conda_binary": "a_conda_binary",
            "cram_reference": "grch37_homo_sapiens_-d5-.fasta",
            "slurm": {
                "account": "development",
                "conda_env": "S_crunchy",
                "mail_user": "an@scilifelab.se",
            },
        },
        "data-delivery": {
            "account": "development",
            "base_path": "/another/path",
            "covid_destination_path": "server.name.se:/another/%s/foldername/",
            "covid_report_path": "/folder_structure/%s/yet_another_folder/filename_%s_data_*.csv",
            "destination_path": "server.name.se:/some",
            "mail_user": "an@email.com",
        },
        "demultiplex": {
            "run_dir": "tests/fixtures/apps/demultiplexing/flowcell-runs",
            "out_dir": "tests/fixtures/apps/demultiplexing/demultiplexed-runs",
            "slurm": {
                "account": "development",
                "mail_user": "an@scilifelab.se",
            },
        },
        "encryption": {"binary_path": "bin/gpg"},
        "external": {
            "caesar": "server.name.se:/path/%s/on/caesar",
            "hasta": "/path/on/hasta/%s",
        },
        "fluffy": {
            "binary_path": "echo",
            "config_path": "fluffy/Config.json",
            "root_dir": str(fluffy_dir),
            "sftp": {
                "user": "sftpuser",
                "password": "sftpassword",
                "host": "sftphost",
                "remote_path": "sftpremotepath",
                "port": 22,
            },
        },
        "genotype": {
            "binary_path": "echo",
            "config_path": "genotype-stage.yaml",
        },
        "gisaid": {
            "binary_path": "/path/to/gisaid_uploader.py",
            "log_dir": "/path/to/log",
            "logwatch_email": "some@email.com",
            "upload_cid": "cid",
            "upload_password": "pass",
            "submitter": "s.submitter",
        },
        "hermes": {"binary_path": "hermes"},
        "housekeeper": {"database": hk_uri, "root": str(housekeeper_dir)},
        "lims": {
            "host": "https://lims.scilifelab.se",
            "password": "password",
            "username": "user",
        },
        "loqusdb": {"binary_path": "loqusdb", "config_path": "loqusdb-stage.yaml"},
        "loqusdb-wes": {"binary_path": "loqusdb", "config_path": "loqusdb-wes-stage.yaml"},
        "loqusdb-somatic": {"binary_path": "loqusdb", "config_path": "loqusdb-somatic-stage.yaml"},
        "loqusdb-tumor": {"binary_path": "loqusdb", "config_path": "loqusdb-tumor-stage.yaml"},
        "microsalt": {
            "binary_path": "echo",
            "conda_binary": "a_conda_binary",
            "conda_env": "S_microSALT",
            "queries_path": Path(microsalt_dir, "queries").as_posix(),
            "root": str(microsalt_dir),
        },
        "mip-rd-dna": {
            "conda_binary": "a_conda_binary",
            "conda_env": "S_mip9.0",
            "mip_config": "mip9.0-dna-stage.yaml",
            "pipeline": "analyse rd_dna",
            "root": str(mip_dir),
            "script": "mip",
        },
        "mip-rd-rna": {
            "conda_binary": "a_conda_binary",
            "conda_env": "S_mip9.0",
            "mip_config": "mip9.0-rna-stage.yaml",
            "pipeline": "analyse rd_rna",
            "root": str(mip_dir),
            "script": "mip",
        },
        "mutacc-auto": {
            "binary_path": "echo",
            "config_path": "mutacc-auto-stage.yaml",
            "padding": 300,
        },
        "mutant": {
            "binary_path": "echo",
            "conda_binary": "a_conda_binary",
            "conda_env": "S_mutant",
            "root": str(mip_dir),
        },
        "pdc": {"binary_path": "/bin/dsmc"},
        "scout": {
            "binary_path": "echo",
            "config_path": "scout-stage.yaml",
        },
        "statina": {
            "api_url": "api_url",
            "auth_path": "auth_path",
            "host": "http://localhost:28002",
            "key": "key",
            "upload_path": "upload_path",
            "user": "user",
        },
        "tar": {"binary_path": "/bin/tar"},
        "trailblazer": {
            "host": "https://trailblazer.scilifelab.se/",
            "service_account": "SERVICE",
            "service_account_auth_file": "trailblazer-auth.json",
        },
        "vogue": {"binary_path": "echo", "config_path": "vogue-stage.yaml"},
    }


@pytest.fixture(name="cg_context")
def fixture_cg_context(
    context_config: dict, base_store: Store, housekeeper_api: MockHousekeeperAPI
) -> CGConfig:
    """Return a cg config."""
    cg_config = CGConfig(**context_config)
    cg_config.status_db_ = base_store
    cg_config.housekeeper_api_ = housekeeper_api
    return cg_config



from pathlib import Path
from typing import List

from click import testing

from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.lims.samplesheet import (
    LimsFlowcellSample,
    LimsFlowcellSampleBcl2Fastq,
    LimsFlowcellSampleDragen,
)
from cg.cli.demultiplex.sample_sheet import create_sheet
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.flow_cell import FlowCell


def test_create_sample_sheet_no_run_parameters(
    cli_runner: testing.CliRunner,
    flow_cell_working_directory_no_run_parameters: Path,
    sample_sheet_context: CGConfig,
    caplog,
    mocker,
):
    # GIVEN a folder with a non existing sample sheet
    flowcell_object: FlowCell = FlowCell(flow_cell_working_directory_no_run_parameters)
    assert flowcell_object.run_parameters_path.exists() is False
    mocker.patch("cg.cli.demultiplex.sample_sheet.flowcell_samples", return_value=[{"sample": 1}])
    demux_api: DemultiplexingAPI = sample_sheet_context.demultiplex_api
    demux_api.run_dir = flow_cell_working_directory_no_run_parameters.parent
    sample_sheet_context.demultiplex_api_ = demux_api

    # WHEN running the create sample sheet command
    result: testing.Result = cli_runner.invoke(
        create_sheet, [flowcell_object.full_name], obj=sample_sheet_context
    )

    # THEN assert it exits with a non zero exit code
    assert result.exit_code != 0
    # THEN assert the correct information is communicated
    assert "Could not find run parameters file" in caplog.text


def test_create_bcl2fastq_sample_sheet(
    cli_runner: testing.CliRunner,
    flow_cell_working_directory: Path,
    sample_sheet_context: CGConfig,
    lims_novaseq_bcl2fastq_samples: List[LimsFlowcellSampleBcl2Fastq],
    mocker,
):
    # GIVEN a flowcell directory with some run parameters
    flowcell: FlowCell = FlowCell(flow_cell_working_directory)
    assert flowcell.run_parameters_path.exists()
    # GIVEN that there is no sample sheet present
    assert not flowcell.sample_sheet_exists()
    mocker.patch(
        "cg.cli.demultiplex.sample_sheet.flowcell_samples",
        return_value=lims_novaseq_bcl2fastq_samples,
    )
    # GIVEN a lims api that returns some samples

    # WHEN creating a sample sheet
    result = cli_runner.invoke(
        create_sheet, [str(flow_cell_working_directory)], obj=sample_sheet_context
    )

    # THEN assert it exits with success
    assert result.exit_code == 0
    # THEN assert that the sample sheet was created
    assert flowcell.sample_sheet_exists()
    # THEN assert that the sample sheet is on the correct format
    assert flowcell.validate_sample_sheet()


def test_create_dragen_sample_sheet(
    cli_runner: testing.CliRunner,
    flow_cell_working_directory: Path,
    sample_sheet_context: CGConfig,
    lims_novaseq_dragen_samples: List[LimsFlowcellSampleDragen],
    mocker,
):
    # GIVEN a flowcell directory with some run parameters
    flowcell: FlowCell = FlowCell(flow_cell_working_directory, bcl_converter="dragen")
    assert flowcell.run_parameters_path.exists()
    # GIVEN that there is no sample sheet present
    assert not flowcell.sample_sheet_exists()
    mocker.patch(
        "cg.cli.demultiplex.sample_sheet.flowcell_samples",
        return_value=lims_novaseq_dragen_samples,
    )
    # GIVEN a lims api that returns some samples

    # WHEN creating a sample sheet
    result = cli_runner.invoke(
        create_sheet, [str(flow_cell_working_directory), "-b", "dragen"], obj=sample_sheet_context
    )

    # THEN assert it exits with success
    assert result.exit_code == 0
    # THEN assert that the sample sheet was created
    assert flowcell.sample_sheet_exists()
    # THEN assert that the sample sheet is on the correct format
    assert flowcell.validate_sample_sheet()



import logging
from typing import List

import click
from cg.apps.environ import environ_email
from cg.cli.workflow.commands import ensure_flowcells_ondisk, link, resolve_compression
from cg.cli.workflow.mip.options import (
    ARGUMENT_CASE_ID,
    EMAIL_OPTION,
    OPTION_BWA_MEM,
    OPTION_DRY,
    OPTION_MIP_DRY_RUN,
    OPTION_PANEL_BED,
    OPTION_SKIP_EVALUATION,
    QOS_OPTION,
    START_AFTER_PROGRAM,
    START_WITH_PROGRAM,
)
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from cg.exc import CgError, DecompressionNeededError, FlowcellsNeededError
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@click.command("config-case")
@OPTION_DRY
@OPTION_PANEL_BED
@ARGUMENT_CASE_ID
@click.pass_obj
def config_case(context: CGConfig, case_id: str, panel_bed: str, dry_run: bool):
    """Generate a config for the case_id"""

    analysis_api: MipAnalysisAPI = context.meta_apis["analysis_api"]
    try:
        analysis_api.verify_case_id_in_statusdb(case_id)
        panel_bed: str = analysis_api.resolve_panel_bed(panel_bed=panel_bed)
        config_data: dict = analysis_api.pedigree_config(case_id=case_id, panel_bed=panel_bed)
    except CgError as error:
        LOG.error(error)
        raise click.Abort()
    if dry_run:
        click.echo(config_data)
        return
    analysis_api.write_pedigree_config(data=config_data, case_id=case_id)


@click.command()
@OPTION_DRY
@ARGUMENT_CASE_ID
@click.pass_obj
def panel(context: CGConfig, case_id: str, dry_run: bool):
    """Write aggregated gene panel file exported from Scout"""

    analysis_api: MipAnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.verify_case_id_in_statusdb(case_id=case_id)

    bed_lines: List[str] = analysis_api.panel(case_id=case_id)
    if dry_run:
        for bed_line in bed_lines:
            click.echo(bed_line)
        return
    analysis_api.write_panel(case_id, bed_lines)


@click.command()
@QOS_OPTION
@EMAIL_OPTION
@START_AFTER_PROGRAM
@START_WITH_PROGRAM
@ARGUMENT_CASE_ID
@OPTION_BWA_MEM
@OPTION_DRY
@OPTION_MIP_DRY_RUN
@OPTION_SKIP_EVALUATION
@click.pass_obj
def run(
    context: CGConfig,
    case_id: str,
    dry_run: bool = False,
    email: str = None,
    mip_dry_run: bool = False,
    slurm_quality_of_service: str = None,
    skip_evaluation: bool = False,
    start_after: str = None,
    start_with: str = None,
    use_bwa_mem: bool = False,
):
    """Run the analysis for a case"""

    analysis_api: MipAnalysisAPI = context.meta_apis["analysis_api"]

    analysis_api.verify_case_id_in_statusdb(case_id)
    command_args = dict(
        slurm_quality_of_service=slurm_quality_of_service
        or analysis_api.get_slurm_qos_for_case(case_id),
        email=email or environ_email(),
        dryrun=mip_dry_run,
        start_after=start_after,
        start_with=start_with,
        skip_evaluation=analysis_api.get_skip_evaluation_flag(
            case_id=case_id, skip_evaluation=skip_evaluation
        ),
        use_bwa_mem=use_bwa_mem,
    )

    try:
        analysis_api.check_analysis_ongoing(case_id=case_id)
    except CgError as error:
        LOG.error(error)
        raise click.Abort
    analysis_api.run_analysis(case_id=case_id, dry_run=dry_run, command_args=command_args)

    if dry_run:
        LOG.info("Running in dry-run mode.")
        return

    if mip_dry_run:
        LOG.info("Executed MIP in dry-run mode")
        return

    try:
        analysis_api.add_pending_trailblazer_analysis(case_id=case_id)
        analysis_api.set_statusdb_action(case_id=case_id, action="running")
        LOG.info("%s run started!", analysis_api.pipeline)
    except CgError as error:
        LOG.error(error)
        raise click.Abort


@click.command()
@ARGUMENT_CASE_ID
@EMAIL_OPTION
@OPTION_BWA_MEM
@OPTION_DRY
@OPTION_MIP_DRY_RUN
@OPTION_PANEL_BED
@OPTION_SKIP_EVALUATION
@QOS_OPTION
@START_AFTER_PROGRAM
@START_WITH_PROGRAM
@click.pass_context
def start(
    context: click.Context,
    case_id: str,
    dry_run: bool,
    email: str,
    mip_dry_run: bool,
    panel_bed: str,
    slurm_quality_of_service: str,
    skip_evaluation: bool,
    start_after: str,
    start_with: str,
    use_bwa_mem: bool,
):
    """Start full MIP analysis workflow for a case"""

    analysis_api: MipAnalysisAPI = context.obj.meta_apis["analysis_api"]

    analysis_api.verify_case_id_in_statusdb(case_id=case_id)
    LOG.info("Starting full MIP analysis workflow for case %s", case_id)
    try:
        context.invoke(ensure_flowcells_ondisk, case_id=case_id)
        context.invoke(resolve_compression, case_id=case_id, dry_run=dry_run)
        context.invoke(link, case_id=case_id, dry_run=dry_run)
        context.invoke(panel, case_id=case_id, dry_run=dry_run)
        context.invoke(config_case, case_id=case_id, panel_bed=panel_bed, dry_run=dry_run)
        context.invoke(
            run,
            case_id=case_id,
            slurm_quality_of_service=slurm_quality_of_service,
            email=email,
            start_after=start_after,
            start_with=start_with,
            dry_run=dry_run,
            mip_dry_run=mip_dry_run,
            skip_evaluation=skip_evaluation,
            use_bwa_mem=use_bwa_mem,
        )
    except (FlowcellsNeededError, DecompressionNeededError) as error:
        LOG.error(error)


@click.command("start-available")
@OPTION_DRY
@click.pass_context
def start_available(context: click.Context, dry_run: bool = False):
    """Start full MIP analysis workflow for all cases ready for analysis"""

    analysis_api: MipAnalysisAPI = context.obj.meta_apis["analysis_api"]

    exit_code: int = EXIT_SUCCESS
    for case_obj in analysis_api.get_cases_to_analyze():
        try:
            context.invoke(start, case_id=case_obj.internal_id, dry_run=dry_run)
        except CgError as error:
            LOG.error(error)
            exit_code = EXIT_FAIL
        except Exception as error:
            LOG.error(f"Unspecified error occurred: %s", error)
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort



"""API for uploading rare disease observations."""

import logging
from typing import Dict

from housekeeper.store.models import Version, File

from cg.apps.loqus import LoqusdbAPI
from cg.constants.observations import (
    MipDNAObservationsAnalysisTag,
    MipDNALoadParameters,
    LoqusdbInstance,
    LOQUSDB_MIP_SEQUENCING_METHODS,
    LOQUSDB_ID,
)
from cg.constants.sequencing import SequencingMethod
from cg.exc import LoqusdbUploadCaseError, LoqusdbDuplicateRecordError, CaseNotFoundError
from cg.meta.observations.observations_api import ObservationsAPI
from cg.models.cg_config import CGConfig
from cg.models.observations.input_files import MipDNAObservationsInputFiles
from cg.store import models
from cg.utils.dict import get_full_path_dictionary

LOG = logging.getLogger(__name__)


class MipDNAObservationsAPI(ObservationsAPI):
    """API to manage MIP-DNA observations."""

    def __init__(self, config: CGConfig, sequencing_method: SequencingMethod):
        super().__init__(config)
        self.sequencing_method: SequencingMethod = sequencing_method
        self.loqusdb_api: LoqusdbAPI = self.get_loqusdb_api(self.get_loqusdb_instance())

    def get_loqusdb_instance(self) -> LoqusdbInstance:
        """Return the Loqusdb instance associated to the sequencing method."""
        if self.sequencing_method not in LOQUSDB_MIP_SEQUENCING_METHODS:
            LOG.error(
                f"Sequencing method {self.sequencing_method} is not supported by Loqusdb. Cancelling upload."
            )
            raise LoqusdbUploadCaseError

        loqusdb_instances: Dict[SequencingMethod, LoqusdbInstance] = {
            SequencingMethod.WGS: LoqusdbInstance.WGS,
            SequencingMethod.WES: LoqusdbInstance.WES,
        }
        return loqusdb_instances[self.sequencing_method]

    def load_observations(
        self, case: models.Family, input_files: MipDNAObservationsInputFiles
    ) -> None:
        """Load observation counts to Loqusdb for a MIP-DNA case."""
        if case.tumour_samples:
            LOG.error(f"Case {case.internal_id} has tumour samples. Cancelling upload.")
            raise LoqusdbUploadCaseError

        if self.is_duplicate(
            case=case,
            loqusdb_api=self.loqusdb_api,
            profile_vcf_path=input_files.profile_vcf_path,
            profile_threshold=MipDNALoadParameters.PROFILE_THRESHOLD.value,
        ):
            LOG.error(
                f"Case {case.internal_id} has already been uploaded to {repr(self.loqusdb_api)}"
            )
            raise LoqusdbDuplicateRecordError

        load_output: dict = self.loqusdb_api.load(
            case_id=case.internal_id,
            snv_vcf_path=input_files.snv_vcf_path,
            sv_vcf_path=input_files.sv_vcf_path,
            profile_vcf_path=input_files.profile_vcf_path,
            family_ped_path=input_files.family_ped_path,
            gq_threshold=MipDNALoadParameters.GQ_THRESHOLD.value,
            hard_threshold=MipDNALoadParameters.HARD_THRESHOLD.value,
            soft_threshold=MipDNALoadParameters.SOFT_THRESHOLD.value,
        )
        loqusdb_id: str = str(self.loqusdb_api.get_case(case_id=case.internal_id)[LOQUSDB_ID])
        self.update_statusdb_loqusdb_id(samples=case.samples, loqusdb_id=loqusdb_id)
        LOG.info(f"Uploaded {load_output['variants']} variants to {repr(self.loqusdb_api)}")

    def extract_observations_files_from_hk(
        self, hk_version: Version
    ) -> MipDNAObservationsInputFiles:
        """Extract observations files given a housekeeper version for rare diseases."""
        input_files: Dict[str, File] = {
            "snv_vcf_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[MipDNAObservationsAnalysisTag.SNV_VCF]
            ).first(),
            "sv_vcf_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[MipDNAObservationsAnalysisTag.SV_VCF]
            ).first()
            if self.sequencing_method == SequencingMethod.WGS
            else None,
            "profile_vcf_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[MipDNAObservationsAnalysisTag.PROFILE_GBCF]
            ).first(),
            "family_ped_path": self.housekeeper_api.files(
                version=hk_version.id, tags=[MipDNAObservationsAnalysisTag.FAMILY_PED]
            ).first(),
        }
        return MipDNAObservationsInputFiles(**get_full_path_dictionary(input_files))

    def delete_case(self, case: models.Family) -> None:
        """Delete rare disease case observations from Loqusdb."""
        if not self.loqusdb_api.get_case(case.internal_id):
            LOG.error(
                f"Case {case.internal_id} could not be found in Loqusdb. Skipping case deletion."
            )
            raise CaseNotFoundError

        self.loqusdb_api.delete_case(case.internal_id)
        self.update_statusdb_loqusdb_id(samples=case.samples, loqusdb_id=None)
        LOG.info(f"Removed observations for case {case.internal_id} from {repr(self.loqusdb_api)}")



"""Test methods for cg cli set list_keys"""
import logging

import pytest
from cg.cli.set.base import list_keys
from cg.constants import EXIT_SUCCESS, Priority
from cg.constants.subject import Gender
from cg.models.cg_config import CGConfig
from cg.store import Store
from click.testing import CliRunner


def test_list_keys_without_sample(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers, caplog
):
    # GIVEN a database with no sample

    # WHEN setting sample but asking for help
    with caplog.at_level(logging.INFO):
        result = cli_runner.invoke(list_keys, obj=base_context)

    # THEN it should not fail on not having a sample as argument
    assert result.exit_code == EXIT_SUCCESS

    # THEN the flags should have been mentioned in the output
    assert "-kv" in caplog.text

    # THEN the name property should have been mentioned
    assert "name" in caplog.text


def test_list_keys_with_sample(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers, caplog
):
    # GIVEN a database with a sample

    sample_obj = helpers.add_sample(base_store, gender=Gender.FEMALE)

    # WHEN setting sample but skipping lims
    with caplog.at_level(logging.INFO):
        result = cli_runner.invoke(
            list_keys, ["--sample_id", sample_obj.internal_id], obj=base_context
        )

    # THEN it should not fail on having a sample as argument
    assert result.exit_code == EXIT_SUCCESS

    # THEN the flags should have been mentioned in the output
    assert "-kv" in caplog.text

    # THEN the name property should have been mentioned
    assert "name" in caplog.text

    # THEN the name value should have been mentioned
    assert sample_obj.name in caplog.text



"""Test for upload API."""
import datetime as dt

from cg.meta.upload.upload_api import UploadAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Family, Analysis
from tests.cli.workflow.conftest import tb_api
from tests.store_helpers import StoreHelpers


def test_update_uploaded_at(dna_mip_context: CGConfig, helpers: StoreHelpers, mip_case: Family):
    """Test setting uploaded at for a finished analysis."""
    # GIVEN an analysis that should be uploaded
    upload_api: UploadAPI = UploadAPI(
        config=dna_mip_context, analysis_api=dna_mip_context.meta_apis["analysis_api"]
    )
    mip_analysis: Analysis = mip_case.analyses[0]
    assert mip_analysis.uploaded_at is None

    # WHEN setting the uploaded at
    upload_api.update_uploaded_at(analysis=mip_analysis)

    # THEN the uploaded at should be set to a date
    assert isinstance(mip_analysis.uploaded_at, dt.datetime)



import http
import os
import tempfile
from datetime import date

from flask import (
    Blueprint,
    current_app,
    flash,
    redirect,
    render_template,
    request,
    send_from_directory,
    session,
    url_for,
    abort,
)
from flask_dance.contrib.google import google

from cg.apps.invoice.render import render_xlsx
from cg.meta.invoice import InvoiceAPI
from cg.server.ext import db, lims

BLUEPRINT = Blueprint("invoices", __name__, template_folder="templates")


def logged_in():
    user_obj = db.user(session.get("user_email"))
    return google.authorized and user_obj and user_obj.is_admin


def undo_invoice(invoice_id):
    invoice_obj = db.invoice(invoice_id)
    record_type = invoice_obj.record_type
    records = db.invoice_samples(invoice_id=invoice_id)
    invoice_obj.delete()
    for record in records:
        record.invoice_id = None
    db.session.commit()

    return url_for(".new", record_type=record_type)


def make_new_invoice():
    customer_id = request.form.get("customer")
    customer_obj = db.customer(customer_id)
    record_ids = request.form.getlist("records")
    record_type = request.form.get("record_type")
    if len(record_ids) == 0:
        return redirect(url_for(".new", record_type=record_type))
    if record_type == "Pool":
        pools = [db.pool(pool_id) for pool_id in record_ids]
        new_invoice = db.add_invoice(
            customer=customer_obj,
            pools=pools,
            comment=request.form.get("comment"),
            discount=int(request.form.get("discount", "0")),
            record_type="Pool",
        )
    elif record_type == "Sample":
        samples = [db.sample(sample_id) for sample_id in record_ids]
        new_invoice = db.add_invoice(
            customer=customer_obj,
            samples=samples,
            comment=request.form.get("comment"),
            discount=int(request.form.get("discount", "0")),
            record_type="Sample",
        )

    db.add_commit(new_invoice)
    return url_for(".invoice", invoice_id=new_invoice.id)


def upload_invoice_news_to_db():
    invoice_id = request.form.get("invoice_id")
    invoice_obj = db.invoice(invoice_id)
    invoice_obj.comment = request.form.get("comment")

    if request.form.get("final_price") != invoice_obj.price:
        invoice_obj.price = request.form.get("final_price")

    if request.form.get("invoice_sent") and not invoice_obj.invoiced_at:
        invoice_obj.invoiced_at = date.today()
    elif not request.form.get("invoice_sent"):
        invoice_obj.invoiced_at = None

    kth_excel_file = request.files.get("KTH_excel")
    if kth_excel_file:
        invoice_obj.excel_kth = kth_excel_file.stream.read()
    ki_excel_file = request.files.get("KI_excel")
    if ki_excel_file:
        invoice_obj.excel_ki = ki_excel_file.stream.read()
    db.commit()
    return url_for("invoices.invoice", invoice_id=invoice_id)


@BLUEPRINT.route("/", methods=["GET", "POST"])
def index():
    """Show invoices."""
    if not logged_in():
        return redirect(url_for("admin.index"))

    if request.form.get("new_invoice_updates"):
        url = upload_invoice_news_to_db()
        return redirect(url)
    elif request.form.get("undo"):
        invoice_id = request.form.get("invoice_id")
        url = undo_invoice(invoice_id)
        return redirect(url)
    elif request.method == "POST":
        url = make_new_invoice()
        return redirect(url)
    else:
        invoices = {
            "sent_invoices": db.invoices(invoiced=True),
            "pending_invoices": db.invoices(invoiced=False),
        }
        return render_template("invoices/index.html", invoices=invoices)


@BLUEPRINT.route("/new/<record_type>")
def new(record_type):
    """Generate a new invoice."""
    if not logged_in():
        return redirect(url_for("admin.index"))

    count = request.args.get("total", 0)
    customer_id = request.args.get("customer", "cust002")
    customer_obj = db.customer(customer_id)

    if record_type == "Sample":
        records, customers_to_invoice = db.samples_to_invoice(customer=customer_obj)
    elif record_type == "Pool":
        records, customers_to_invoice = db.pools_to_invoice(customer=customer_obj)
    return render_template(
        "invoices/new.html",
        customers_to_invoice=customers_to_invoice,
        count=count,
        records=records,
        record_type=record_type,
        total_price_treshold=current_app.config["TOTAL_PRICE_TRESHOLD"],
        args={"customer": customer_id},
    )


@BLUEPRINT.route("/<int:invoice_id>", methods=["GET", "POST"])
def invoice(invoice_id):
    """Save comments and uploaded modified invoices."""
    if not logged_in():
        return redirect(url_for("admin.index"))

    invoice_obj = db.invoice(invoice_id)
    api = InvoiceAPI(db, lims, invoice_obj)
    kth_inv = api.prepare("KTH")
    ki_inv = api.prepare("KI")

    if not (kth_inv and ki_inv):
        flash(" ,".join(list(set(api.log))))
        undo_invoice(invoice_id)
        return redirect(request.referrer)

    if not invoice_obj.price:
        final_price = api.total_price()
    else:
        final_price = invoice_obj.price

    return render_template(
        "invoices/invoice.html",
        invoice=invoice_obj,
        invoice_dict={"KTH": kth_inv, "KI": ki_inv},
        default_price=api.total_price(),
        final_price=final_price,
        record_type=invoice_obj.record_type,
    )


@BLUEPRINT.route("/<int:invoice_id>/excel")
def invoice_template(invoice_id):
    """Generate invoice template"""
    if not logged_in():
        return redirect(url_for("admin.index"))

    cost_center = request.args.get("cost_center", "KTH")
    invoice_obj = db.invoice(invoice_id)
    api = InvoiceAPI(db, lims, invoice_obj)

    invoice_dict = api.prepare(cost_center)
    workbook = render_xlsx(invoice_dict)

    temp_dir = tempfile.gettempdir()
    filename = "Invoice_{}_{}.xlsx".format(invoice_obj.id, cost_center)
    excel_path = os.path.join(temp_dir, filename)
    workbook.save(excel_path)

    return send_from_directory(directory=temp_dir, filename=filename, as_attachment=True)


@BLUEPRINT.route("/<int:invoice_id>/invoice_file/<cost_center>")
def modified_invoice(invoice_id, cost_center):
    """Enables download of modified invoices saved in the database."""
    if not logged_in():
        return redirect(url_for("admin.index"))

    if cost_center not in ["KTH", "KI"]:
        return abort(http.HTTPStatus.BAD_REQUEST)

    invoice_obj = db.invoice(invoice_id)
    file_name = "invoice_" + cost_center + str(invoice_id) + ".xlsx"
    temp_dir = tempfile.mkdtemp()
    file_path = os.path.join(temp_dir, file_name)

    with open(file_path, "wb") as file_object:
        if cost_center == "KTH":
            file_object.write(invoice_obj.excel_kth)
        elif cost_center == "KI":
            file_object.write(invoice_obj.excel_ki)
        pass

    return send_from_directory(directory=temp_dir, filename=file_name, as_attachment=True)



"""This script tests the cli methods to get samples in status-db"""

from cg.cli.get import get
from cg.models.cg_config import CGConfig
from cg.store import Store
from click.testing import CliRunner
from tests.store_helpers import StoreHelpers


def test_get_sample_bad_sample(cli_runner: CliRunner, base_context: CGConfig):
    """Test to get a sample using a non-existing sample-id"""
    # GIVEN an empty database

    # WHEN getting a sample
    name = "dummy_name"
    result = cli_runner.invoke(get, ["sample", name], obj=base_context)

    # THEN it should warn about missing sample id instead of getting a sample
    # it will not fail since the API accepts multiple samples
    assert result.exit_code == 0


def test_get_sample_required(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test to get a sample using only the required argument"""
    # GIVEN a database with a sample
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    assert disk_store.Sample.query.count() == 1

    # WHEN getting a sample
    result = cli_runner.invoke(get, ["sample", sample_id], obj=base_context)

    # THEN it should have been get
    assert result.exit_code == 0
    assert sample_id in result.output


def test_get_samples_required(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test to get several samples using only the required arguments"""
    # GIVEN a database with two samples
    sample_id1 = "1"
    helpers.add_sample(store=disk_store, internal_id=sample_id1)
    sample_id2 = "2"
    helpers.add_sample(store=disk_store, internal_id=sample_id2)

    assert disk_store.Sample.query.count() == 2

    # WHEN getting a sample
    result = cli_runner.invoke(get, ["sample", sample_id1, sample_id2], obj=base_context)

    # THEN it should have been get
    assert result.exit_code == 0
    assert sample_id1 in result.output
    assert sample_id2 in result.output


def test_get_sample_output(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output has the data of the sample"""
    # GIVEN a database with a sample with data
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    name = sample.name
    customer_id = sample.customer.internal_id
    application_tag = sample.application_version.application.tag
    state = sample.state
    priority_human = sample.priority_human

    # WHEN getting a sample
    result = cli_runner.invoke(get, ["sample", sample_id], obj=base_context)

    # THEN it should have been get
    assert result.exit_code == 0
    assert sample_id in result.output
    assert name in result.output
    assert customer_id in result.output
    assert application_tag in result.output
    assert state in result.output
    assert priority_human in result.output


def test_get_sample_external_false(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output has the external-value of the sample"""
    # GIVEN a database with a sample with data
    sample = helpers.add_sample(disk_store, is_external=False)
    sample_id = sample.internal_id
    is_external_false = "No"
    is_external_true = "Yes"

    # WHEN getting a sample
    result = cli_runner.invoke(get, ["sample", sample_id], obj=base_context)

    # THEN it should have been get
    assert result.exit_code == 0
    assert is_external_false in result.output
    assert is_external_true not in result.output


def test_get_sample_external_true(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output has the external-value of the sample"""
    # GIVEN a database with a sample with data
    sample = helpers.add_sample(disk_store, is_external=True)
    sample_id = sample.internal_id
    is_external_false = "No"
    is_external_true = "Yes"

    # WHEN getting a sample
    result = cli_runner.invoke(get, ["sample", sample_id], obj=base_context)

    # THEN it should have been get
    assert result.exit_code == 0
    assert is_external_true in result.output
    assert is_external_false not in result.output


def test_get_sample_no_families_without_family(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the --no-families flag works without families"""
    # GIVEN a database with a sample without related samples
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    assert not disk_store.Sample.query.first().links

    # WHEN getting a sample with the --no-families flag
    result = cli_runner.invoke(get, ["sample", sample_id, "--no-families"], obj=base_context)

    # THEN everything is fine
    assert result.exit_code == 0


def test_get_sample_no_families_with_family(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the --no-families flag doesn't show case info"""
    # GIVEN a database with a sample with related samples
    case = helpers.add_case(disk_store)
    sample = helpers.add_sample(disk_store)
    link = helpers.add_relationship(disk_store, sample=sample, case=case)
    assert link in disk_store.Sample.query.first().links
    sample_id = sample.internal_id

    # WHEN getting a sample with the --no-families flag
    result = cli_runner.invoke(get, ["sample", sample_id, "--no-families"], obj=base_context)

    # THEN all related families should be listed in the output
    assert result.exit_code == 0
    for link in disk_store.Sample.query.first().links:
        assert link.family.internal_id not in result.output


def test_get_sample_families_without_family(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the --families flag works without families"""
    # GIVEN a database with a sample without related samples
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    assert not disk_store.Sample.query.first().links

    # WHEN getting a sample with the --families flag
    result = cli_runner.invoke(get, ["sample", sample_id, "--families"], obj=base_context)

    # THEN everything is fine
    assert result.exit_code == 0


def test_get_sample_families_with_family(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the --families flag does show case info"""
    # GIVEN a database with a sample with related samples
    case = helpers.add_case(disk_store)
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    helpers.add_relationship(disk_store, sample=sample, case=case)
    assert disk_store.Sample.query.first().links

    # WHEN getting a sample with the --families flag
    result = cli_runner.invoke(get, ["sample", sample_id, "--families"], obj=base_context)

    # THEN all related families should be listed in the output
    assert result.exit_code == 0
    for link in disk_store.Sample.query.first().links:
        assert link.family.internal_id in result.output


def test_hide_sample_flowcells_without_flowcell(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that we can query samples and hide flow cell even when there are none."""
    # GIVEN a database with a sample without related flow cell
    sample = helpers.add_sample(disk_store)
    sample_id = sample.internal_id
    assert not disk_store.Flowcell.query.first()

    # WHEN getting a sample with the --flowcells flag
    result = cli_runner.invoke(get, ["sample", sample_id, "--hide-flowcell"], obj=base_context)

    # THEN everything is fine
    assert result.exit_code == 0


def test_get_sample_flowcells_with_flowcell(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that we can query samples and hide flow cell, ensuring that no flow cell name is in the output."""
    # GIVEN a database with a sample and a related flow cell
    flow_cell = helpers.add_flowcell(disk_store)
    sample = helpers.add_sample(disk_store, flowcell=flow_cell)
    assert flow_cell in disk_store.Sample.query.first().flowcells
    sample_id = sample.internal_id

    # WHEN getting a sample with the --flowcells flag
    result = cli_runner.invoke(get, ["sample", sample_id, "--hide-flowcell"], obj=base_context)

    # THEN the related flow cell should be listed in the output
    assert result.exit_code == 0
    for flow_cell in disk_store.Sample.query.first().flowcells:
        assert flow_cell.name not in result.output



import hashlib
import logging
from pathlib import Path

LOG = logging.getLogger(__name__)


def check_md5sum(file_path: Path, md5sum: str) -> bool:
    """Checks if the given md5_sum matches that of the given file"""
    BYTES_PER_CHUNK: int = 8192
    with open(file_path, "rb") as file:
        file_hash = hashlib.md5()
        chunk = file.read(BYTES_PER_CHUNK)
        while chunk:
            file_hash.update(chunk)
            chunk = file.read(BYTES_PER_CHUNK)
    calculated_md5sum = file_hash.hexdigest()
    if md5sum == calculated_md5sum:
        return True
    LOG.info("The given md5sum does not match the md5sum for file %s" % file_path)
    return False


def extract_md5sum(md5sum_file: Path) -> str:
    """Searches the first line of a file for a string matching the length of an md5sum"""
    with open(md5sum_file, "r") as md_file:
        first_line: str = md_file.readline()
    line_list = first_line.split(" ")
    for element in line_list:
        element = element.strip()
        if len(element) == 32 and "." not in element:
            md5sum = element
            return md5sum
    LOG.info("No valid md5sum found in file {}.".format(md5sum_file))
    return ""


def sha512_checksum(file: Path) -> str:
    """Generates the sha512 checksum of a file"""
    LOG.debug("Checksum for file %s: ", file)
    sha512_hash = hashlib.sha512()
    with open(file, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            sha512_hash.update(chunk)
    LOG.debug("Result: %s", sha512_hash.hexdigest())
    return sha512_hash.hexdigest()



"""API for tar archiving and extraction on Hasta"""
import logging
from pathlib import Path
from typing import List

from cg.constants.extraction import FlowCellExtractionParameters
from cg.utils import Process

LOG = logging.getLogger(__name__)


class TarAPI:
    """Class that uses tar for various archiving and archive extraction functionality"""

    def __init__(self, binary_path: str, dry_run: bool = False):
        self.process: Process = Process(binary=binary_path)
        self.dry_run: bool = dry_run

    def run_tar_command(self, command: list) -> None:
        """Runs a Tar command"""
        LOG.info("Starting Tar command:")
        LOG.info(f"{self.process.binary} {' '.join(command)}")
        self.process.run_command(command, dry_run=self.dry_run)

    @staticmethod
    def get_extract_file_command(input_file: Path, output_dir: Path) -> List[str]:
        """Generates the Tar command for flow cel run directory extraction"""
        extraction_parameters: list = FlowCellExtractionParameters.EXTRACT_FILE.copy()
        extraction_parameters.append(str(input_file))
        exclude_files: list = FlowCellExtractionParameters.EXCLUDE_FILES.copy()
        extraction_parameters.extend(exclude_files)
        target_directory_parameters: list = FlowCellExtractionParameters.CHANGE_TO_DIR.copy()
        extraction_parameters.extend(target_directory_parameters)
        extraction_parameters.append(str(output_dir))
        return extraction_parameters



"""Tests for the compress fastq cli"""

import datetime as dt
import logging

from cg.cli.compress.fastq import fastq_cmd
from cg.constants import Pipeline
from cg.models.cg_config import CGConfig
from click.testing import CliRunner
from cg.store import Store

from tests.store_helpers import StoreHelpers


def test_compress_fastq_cli_no_family(compress_context: CGConfig, cli_runner: CliRunner, caplog):
    """Test to run the compress command with a database without samples"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context without families

    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, [], obj=compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert "Individuals in 0 (completed) cases where compressed" in caplog.text


def test_compress_fastq_cli_case_id_no_family(
    compress_context: CGConfig, cli_runner: CliRunner, caplog
):
    """Test to run the compress command when no families are found"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context without families
    case_id = "notrealcase"
    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, ["--case-id", case_id], obj=compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert f"Could not find case {case_id}" in caplog.text


def test_compress_fastq_cli_case_id(
    populated_compress_context: CGConfig, cli_runner: CliRunner, helpers: StoreHelpers, caplog
):
    """Test to run the compress command with a specified case id"""
    caplog.set_level(logging.DEBUG)
    status_db: Store = populated_compress_context.status_db

    # GIVEN a context with a case that can be compressed
    case_id = "chonkywombat"

    valid_compressable_case = helpers.add_case(
        store=status_db,
        name=case_id,
        internal_id=case_id,
        data_analysis=Pipeline.MIP_DNA,
        action=None,
    )
    valid_compressable_case.created_at = dt.datetime.now() - dt.timedelta(days=1000)
    sample1 = helpers.add_sample(store=status_db, internal_id="ACCR9000")
    sample2 = helpers.add_sample(store=status_db, internal_id="ACCR9001")
    helpers.add_relationship(
        store=status_db,
        sample=sample1,
        case=valid_compressable_case,
    )
    helpers.add_relationship(
        store=status_db,
        sample=sample2,
        case=valid_compressable_case,
    )
    status_db.commit()

    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, ["--case-id", case_id], obj=populated_compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert f"Individuals in 1 (completed) cases where compressed" in caplog.text


def test_compress_fastq_cli_multiple_family(
    populated_multiple_compress_context: CGConfig, cli_runner: CliRunner, caplog
):
    """Test to run the compress command with multiple families"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a database with multiple families
    nr_cases = populated_multiple_compress_context.status_db.families().count()
    assert nr_cases > 1

    # WHEN running the compress command
    res = cli_runner.invoke(
        fastq_cmd, ["--number-of-conversions", nr_cases], obj=populated_multiple_compress_context
    )

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert f"Individuals in {nr_cases} (completed) cases where compressed" in caplog.text


def test_compress_fastq_cli_multiple_set_limit(
    populated_multiple_compress_context: CGConfig, cli_runner: CliRunner, caplog
):
    """Test to run the compress command with multiple families and use a limit"""
    compress_context = populated_multiple_compress_context
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with more families than the limit
    nr_cases = compress_context.status_db.families().count()
    limit = 5
    assert nr_cases > limit

    # WHEN running the compress command
    res = cli_runner.invoke(fastq_cmd, ["--number-of-conversions", limit], obj=compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated no more than the limited number of cases was compressed
    assert f"Individuals in {limit} (completed) cases where compressed" in caplog.text



"""Test how the api handles files."""
from datetime import datetime
from pathlib import Path
from typing import List

from cg.apps.housekeeper.hk import HousekeeperAPI
from tests.mocks.hk_mock import MockHousekeeperAPI

from housekeeper.store import models as hk_models

from tests.small_helpers import SmallHelpers


def test_new_file(bed_file: Path, housekeeper_api: MockHousekeeperAPI, small_helpers: SmallHelpers):
    """Test to create a new file with the Housekeeper API"""
    # GIVEN a housekeeper api without files and the path to an existing file
    assert small_helpers.length_of_iterable(housekeeper_api.files()) == 0
    assert bed_file.exists() is True

    # WHEN creating a new file
    new_file_obj = housekeeper_api.new_file(path=bed_file.as_posix())

    # THEN assert a file object was created
    assert new_file_obj

    # THEN assert that the path is correct
    assert new_file_obj.path == bed_file.as_posix()

    # THEN assert that no file is added to the database
    assert small_helpers.length_of_iterable(housekeeper_api.files()) == 0


def test_new_file_non_existing_path(housekeeper_api: MockHousekeeperAPI):
    """Test to create a new file with the Housekeeper API."""
    # GIVEN a housekeeper api without files and the path to a not existing file
    file_name = Path("a_file.hello")
    assert file_name.exists() is False

    # WHEN creating a new file
    new_file_obj = housekeeper_api.new_file(path=file_name.as_posix())

    # THEN assert a file object was created
    assert new_file_obj

    # THEN assert that the path is correct
    assert new_file_obj.path == file_name.as_posix()


def test_add_new_file(
    populated_housekeeper_api: MockHousekeeperAPI,
    case_id: str,
    madeline_output: Path,
    small_helpers: SmallHelpers,
    not_existing_hk_tag: str,
):
    """Test to create a new file with the Housekeeper API."""
    # GIVEN a populated housekeeper api and the existing version of a bundle
    version_obj = populated_housekeeper_api.last_version(bundle=case_id)

    # GIVEN an existing file that is not included in the database
    assert madeline_output.exists() is True

    # GIVEN a tag that does not exist
    assert populated_housekeeper_api.tag(name=not_existing_hk_tag) is None

    # GIVEN a known number of files in the db
    nr_files_in_db = small_helpers.length_of_iterable(populated_housekeeper_api.files())

    # WHEN creating a new file
    new_file_obj = populated_housekeeper_api.add_file(
        path=madeline_output, version_obj=version_obj, tags=not_existing_hk_tag
    )

    # THEN assert a file object was created
    assert new_file_obj

    # THEN assert that the path is correct
    assert new_file_obj.path == madeline_output.resolve().as_posix()

    # THEN assert that the file was not added to the database
    new_nr_files = small_helpers.length_of_iterable(populated_housekeeper_api.files())
    assert new_nr_files == nr_files_in_db + 1


def test_get_files(
    populated_housekeeper_api: MockHousekeeperAPI,
    case_id: str,
    tags: List[str],
    small_helpers: SmallHelpers,
):
    """Test get files method."""
    # GIVEN a populated housekeeper api with some files
    nr_files = small_helpers.length_of_iterable(populated_housekeeper_api.files())
    assert nr_files > 0

    # WHEN fetching all files
    files = populated_housekeeper_api.get_files(bundle=case_id, tags=tags)

    # THEN assert all files where fetched
    assert small_helpers.length_of_iterable(files) == nr_files


def test_get_file(populated_housekeeper_api: MockHousekeeperAPI):
    """Test to get a file from the database."""
    # GIVEN a housekeeper api with a file
    file_obj = populated_housekeeper_api.files().first()
    assert file_obj

    # GIVEN the id of a file that exists in HK
    file_id = file_obj.id

    # WHEN fetching the file with get_file
    hk_file = populated_housekeeper_api.get_file(file_id)

    # THEN assert a file was returned
    assert hk_file is not None


def test_delete_file(populated_housekeeper_api: HousekeeperAPI):
    """Test to delete a file from the database."""
    # GIVEN a housekeeper api with a file
    file_obj = populated_housekeeper_api.files().first()
    assert file_obj

    # GIVEN the id of a file that exists in HK
    file_id = file_obj.id

    # WHEN deleting the file
    populated_housekeeper_api.delete_file(file_id)

    # THEN assert the file was removed
    assert populated_housekeeper_api.get_file(file_id) is None


def test_get_included_path(populated_housekeeper_api: MockHousekeeperAPI, case_id: str):
    """Test to get the included path for a file."""
    # GIVEN a populated housekeeper api and the root dir
    root_dir = Path(populated_housekeeper_api.get_root_dir())

    # GIVEN a version and a file object
    version_obj = populated_housekeeper_api.last_version(case_id)
    file_obj = version_obj.files[0]

    # WHEN fetching the included path
    included_path = populated_housekeeper_api.get_included_path(
        root_dir=root_dir, version_obj=version_obj, file_obj=file_obj
    )

    # THEN assert that there is no file existing in the included path
    assert included_path.exists() is False

    # THEN assert that the correct path was created
    assert included_path == Path(root_dir, version_obj.relative_root_dir, Path(file_obj.path).name)


def test_get_include_file(populated_housekeeper_api: MockHousekeeperAPI, case_id: str):
    """Test to get the included path for a file."""
    # GIVEN a populated housekeeper api and the root dir
    root_dir = Path(populated_housekeeper_api.get_root_dir())
    version_obj = populated_housekeeper_api.last_version(case_id)
    file_obj = version_obj.files[0]
    original_path = Path(file_obj.path)
    included_path = Path(root_dir, version_obj.relative_root_dir, original_path.name)

    # GIVEN that the included file does not exist
    assert included_path.exists() is False

    # WHEN including the file
    included_file = populated_housekeeper_api.include_file(file_obj, version_obj)

    # THEN assert that the file has been linked to the included place
    assert included_path.exists() is True

    # THEN assert that the file path has been updated
    assert included_file.path != original_path


def test_check_bundle_files(
    case_id: str,
    timestamp: datetime,
    populated_housekeeper_api: MockHousekeeperAPI,
    hk_version_obj: hk_models.Version,
    fastq_file: Path,
    sample_id: str,
    bed_file: Path,
):
    """Test to see if the function correctly identifies a file that is present and returns a lis without it."""
    # GIVEN a housekeeper version with a file
    version = populated_housekeeper_api.version(bundle=case_id, date=timestamp)

    # WHEN attempting to add two files, one existing and one new
    files_to_add: List[Path] = populated_housekeeper_api.check_bundle_files(
        file_paths=[bed_file, fastq_file],
        bundle_name=case_id,
        last_version=version,
    )

    # Then only the new file should be returned
    assert files_to_add == [fastq_file]


def test_get_tag_names_from_file(populated_housekeeper_api: MockHousekeeperAPI):
    """Test get tag names on a file."""
    # GIVEN a housekeeper api with a file
    file_obj = populated_housekeeper_api.files().first()
    assert file_obj.tags

    # WHEN fetching tags of a file
    tag_names = populated_housekeeper_api.get_tag_names_from_file(file_obj)

    # THEN a list of tag names is returned
    assert tag_names is not None

    # THEN the return type is a list of strings
    assert isinstance(tag_names, list)
    assert all(isinstance(elem, str) for elem in tag_names)



from typing import List

import pytest

from cg.constants import Pipeline
from cg.meta.orders.lims import build_lims_sample
from cg.models.lims.sample import LimsSample
from cg.models.orders.constants import OrderType
from cg.models.orders.order import OrderIn


def test_to_lims_mip(mip_order_to_submit):
    # GIVEN a scout order for a trio
    order_data = OrderIn.parse_obj(obj=mip_order_to_submit, project=OrderType.MIP_DNA)
    # WHEN parsing the order to format for LIMS import
    samples: List[LimsSample] = build_lims_sample(customer="cust003", samples=order_data.samples)

    # THEN it should list all samples
    assert len(samples) == 4

    # THEN container should be 96 well plate for all samples
    assert {sample.container for sample in samples} == {"96 well plate"}

    # THEN container names should be the same for all samples
    container_names = {sample.container_name for sample in samples if sample.container_name}
    assert container_names == {"CMMS"}

    # ... and pick out relevant UDFs
    first_sample: LimsSample = samples[0]
    assert first_sample.well_position == "A:1"
    assert first_sample.udfs.family_name == "family1"
    assert first_sample.udfs.priority == "standard"
    assert first_sample.udfs.application == "WGSPCFC030"
    assert first_sample.udfs.source == "tissue (fresh frozen)"
    assert first_sample.udfs.quantity == "220"
    assert first_sample.udfs.customer == "cust003"
    assert first_sample.udfs.volume == "1.0"

    # THEN assert that the comment of a sample is a string
    assert isinstance(samples[1].udfs.comment, str)


def test_to_lims_fastq(fastq_order_to_submit):
    # GIVEN a fastq order for two samples; normal vs. tumour
    order_data = OrderIn.parse_obj(obj=fastq_order_to_submit, project=OrderType.FASTQ)

    # WHEN parsing the order to format for LIMS
    samples: List[LimsSample] = build_lims_sample(customer="dummyCust", samples=order_data.samples)

    # THEN should "work"
    assert len(samples) == 2
    normal_sample = samples[0]
    tumor_sample = samples[1]
    # ... and pick out relevant UDF values
    assert normal_sample.udfs.tumour is False
    assert tumor_sample.udfs.tumour is True
    assert normal_sample.udfs.volume == "1"


def test_to_lims_rml(rml_order_to_submit):
    # GIVEN a rml order for four samples
    order_data = OrderIn.parse_obj(obj=rml_order_to_submit, project=OrderType.RML)

    # WHEN parsing for LIMS
    samples: List[LimsSample] = build_lims_sample(customer="dummyCust", samples=order_data.samples)

    # THEN it should have found the same number of samples
    assert len(samples) == 4
    # ... and pick out relevant UDFs
    first_sample = samples[0]
    assert first_sample.udfs.pool == "pool-1"
    assert first_sample.udfs.volume == "30"
    assert first_sample.udfs.concentration == "5.0"
    assert first_sample.udfs.index == "IDT DupSeq 10 bp Set B"
    assert first_sample.udfs.index_number == "1"


def test_to_lims_microbial(microbial_order_to_submit):
    # GIVEN a microbial order for three samples
    order_data = OrderIn.parse_obj(obj=microbial_order_to_submit, project=OrderType.MICROSALT)

    # WHEN parsing for LIMS
    samples: List[LimsSample] = build_lims_sample(customer="cust000", samples=order_data.samples)
    # THEN it should "work"

    assert len(samples) == 5
    # ... and pick out relevant UDFs
    first_sample = samples[0].dict()
    assert first_sample["udfs"]["priority"] == "research"
    assert first_sample["udfs"]["organism"] == "M.upium"
    assert first_sample["udfs"]["reference_genome"] == "NC_111"
    assert (
        first_sample["udfs"]["extraction_method"] == "MagNaPure 96 (contact Clinical Genomics "
        "before submission)"
    )
    assert first_sample["udfs"]["volume"] == "1"


def test_to_lims_sarscov2(sarscov2_order_to_submit):
    # GIVEN a sarscov2 order for samples
    order_data = OrderIn.parse_obj(obj=sarscov2_order_to_submit, project=OrderType.SARS_COV_2)

    # WHEN parsing for LIMS
    samples: List[LimsSample] = build_lims_sample(customer="cust000", samples=order_data.samples)

    # THEN it should have found the same number of samples
    assert len(samples) == 5
    # ... and pick out relevant UDFs
    first_sample = samples[0].dict()
    assert first_sample["udfs"]["collection_date"] == "2021-05-05"
    assert first_sample["udfs"]["extraction_method"] == "MagNaPure 96"
    assert first_sample["udfs"]["lab_code"] == "SE110 Växjö"
    assert first_sample["udfs"]["organism"] == "SARS CoV-2"
    assert first_sample["udfs"]["original_lab"] == "Karolinska University Hospital Solna"
    assert first_sample["udfs"]["original_lab_address"] == "171 76 Stockholm"
    assert first_sample["udfs"]["pre_processing_method"] == "COVIDSeq"
    assert first_sample["udfs"]["priority"] == "research"
    assert first_sample["udfs"]["reference_genome"] == "NC_111"
    assert first_sample["udfs"]["region"] == "Stockholm"
    assert first_sample["udfs"]["region_code"] == "01"
    assert first_sample["udfs"]["selection_criteria"] == "1. Allmän övervakning"
    assert first_sample["udfs"]["volume"] == "1"


@pytest.mark.parametrize(
    "project", [OrderType.BALSAMIC, OrderType.BALSAMIC_UMI, OrderType.BALSAMIC_QC]
)
def test_to_lims_balsamic(balsamic_order_to_submit, project):

    # GIVEN a cancer order for a sample
    order_data = OrderIn.parse_obj(obj=balsamic_order_to_submit, project=project)

    # WHEN parsing the order to format for LIMS import
    samples: List[LimsSample] = build_lims_sample(customer="cust000", samples=order_data.samples)
    # THEN it should list all samples

    assert len(samples) == 1
    # ... and determine the container, container name, and well position

    container_names = {sample.container_name for sample in samples if sample.container_name}

    # ... and pick out relevant UDFs
    first_sample = samples[0].dict()
    assert first_sample["name"] == "s1"
    assert {sample.container for sample in samples} == set(["96 well plate"])
    assert first_sample["udfs"]["data_analysis"] in [
        str(Pipeline.BALSAMIC),
        str(Pipeline.BALSAMIC_QC),
        str(Pipeline.BALSAMIC_UMI),
    ]
    assert first_sample["udfs"]["application"] == "WGSPCFC030"
    assert first_sample["udfs"]["sex"] == "M"
    assert first_sample["udfs"]["family_name"] == "family1"
    assert first_sample["udfs"]["customer"] == "cust000"
    assert first_sample["udfs"]["source"] == "blood"
    assert first_sample["udfs"]["volume"] == "1.0"
    assert first_sample["udfs"]["priority"] == "standard"

    assert container_names == set(["p1"])
    assert first_sample["well_position"] == "A:1"
    assert first_sample["udfs"]["tumour"] is True
    assert first_sample["udfs"]["capture_kit"] == "other"
    assert first_sample["udfs"]["tumour_purity"] == "75"

    assert first_sample["udfs"]["formalin_fixation_time"] == "1"
    assert first_sample["udfs"]["post_formalin_fixation_time"] == "2"
    assert first_sample["udfs"]["tissue_block_size"] == "small"

    assert first_sample["udfs"]["quantity"] == "2"
    assert first_sample["udfs"]["comment"] == "other Elution buffer"



"""
    Cg Exceptions.
"""


class CgError(Exception):

    """
    Base exception for the package.
    """

    def __init__(self, message: str = ""):
        super().__init__(message)


class AccessionNumerMissingError(CgError):
    """Raised when accession numers are not found in a gisaid cli log."""


class AnalysisDuplicationError(CgError):
    """
    Error related to trying to create analysis object that already exists in status-db.
    """


class AnalysisNotFinishedError(CgError):
    """
    Exception raised when an adding a MIP analysis to Housekeeper, but the analysis is not
    finished in MIP, as indicated in the qc sample info file.
    """


class AnalysisUploadError(CgError):
    """
    Error related to trying to upload analysis data.
    """


class AnalysisAlreadyUploadedError(CgError):
    """
    Error related to trying to upload an already (or in the process) uploaded analysis.
    """


class BalsamicStartError(CgError):
    """
    Exception raised when Balsamic fails to start.
    """


class BundleAlreadyAddedError(CgError):
    """
    Exception raised when a bundle has already been added to Housekeeper.
    """


class CaseNotFoundError(CgError):
    """
    Exception raised when a case is not found.
    """


class CgDataError(CgError):
    """
    Error related to missing/incomplete data in Status DB.
    """


class ChecksumFailedError(CgError):
    """
    Exception raised when the checksums of two files are not equal.
    """


class DecompressionNeededError(CgError):
    """Raised when decompression still needed to start analysis."""


class DeliveryReportError(CgError):
    """
    Exception related to delivery report creation.
    """


class EmailNotSentError(CgError):
    """Raised when email not sent."""


class FamilyLinkMissingError(CgError):
    """Raised when faimly link missing for a sample."""


class FastaSequenceMissingError(CgError):
    """
    Exception raised when expected sequence in fasta file is missing.
    """


class FlowCellError(CgError):
    """Raised when there is a problem with demultiplexing a flow cell."""


class FlowcellsNeededError(CgError):
    """Raised when fetching flowcells still needed to start analysis."""


class GisaidUploadFailedError(CgError):
    """Raised when gisaid upload fails."""


class HousekeeperFileMissingError(CgError):
    """
    Exception raised when a file is missing in Housekeeper.
    """

    def __init__(self, message: str = "", errors=None):
        super().__init__(message)
        self.errors = errors


class HousekeeperBundleVersionMissingError(CgError):
    """
    Exception raised when bundle version is missing in Housekeeper.
    """


class InvalidFastaError(CgError):
    """
    Exception raised when fasta file content is invalid.
    """


class LimsDataError(CgError):
    """
    Error related to missing/incomplete data in LIMS.
    """


class MandatoryFilesMissing(CgError):
    """
    Exception raised when mandatory files are missing from the deliverables when storing an
    analysis in Housekeeper.
    """


class MipStartError(CgError):
    """
    Exception raised when MIP fails to start a run.
    """


class MultipleFamilyLinksError(CgError):
    """Raised when only one family was expected but more than one was found."""


class OrderError(CgError):
    """
    Exception related to orders.
    """


class OrderFormError(CgError):
    """
    Exception related to the order form.
    """


class PedigreeConfigError(CgError):
    """
    Raised when MIP pedigree config validation fails.
    """

    def __init__(self, message: str = "", errors=None):
        super().__init__(message)
        self.errors = errors


class PipelineUnknownError(CgError):
    """
    Exception raised when a sample in a case has no data analysis type.
    """


class ScoutUploadError(CgError):
    """Raised when uploading to Scout fails."""


class StatinaAPIHTTPError(CgError):
    """Raised when Statina REST API response code is not 200."""


class StoreError(CgError):
    """
    Exception related to storing an analysis.
    """


class TicketCreationError(CgError):
    """
    Exception related to ticket creation.
    """


class TrailblazerAPIHTTPError(CgError):
    """Raised when Trailblazer REST API response code is not 200."""


class TrailblazerMissingAnalysisError(CgError):
    """Raised when Trailblazer REST API response code is not 200."""


class ValidationError(CgError):
    """
    Exception related to delivery report validation.
    """


class DeleteDemuxError(CgError):
    """Raised when there is an issue with wiping a flowcell before start."""


class CustomerPermissionError(CgError):
    """Exception related to the limited permissions of a customer."""


class DataIntegrityError(CgError):
    """Raised when data integrity is not met."""


class LoqusdbError(CgError):
    """Exception related to the Loqusdb app."""


class LoqusdbUploadCaseError(LoqusdbError):
    """Exception raised when a case could not be uploaded to Loqusdb."""


class LoqusdbDeleteCaseError(LoqusdbError):
    """Exception raised when a case cannot be deleted from Loqusdb."""


class LoqusdbDuplicateRecordError(LoqusdbError):
    """Exception related to duplicate records in Loqusdb."""



from enum import Enum
from typing import List, Optional

from cg.constants import DataDelivery, Pipeline
from pydantic import BaseModel, constr, NonNegativeInt, validator

from cg.store import models


class ControlEnum(str, Enum):
    not_control = ""
    positive = "positive"
    negative = "negative"


class SexEnum(str, Enum):
    male = "male"
    female = "female"
    unknown = "unknown"


class PriorityEnum(str, Enum):
    research = "research"
    standard = "standard"
    priority = "priority"
    express = "express"
    clinical_trials = "clinical_trials"


class ContainerEnum(str, Enum):
    no_container = "No container"
    plate = "96 well plate"
    tube = "Tube"


class StatusEnum(str, Enum):
    affected = "affected"
    unaffected = "unaffected"
    unknown = "unknown"


NAME_PATTERN = r"^[A-Za-z0-9-]*$"


class OrderSample(BaseModel):
    age_at_sampling: Optional[str]
    application: constr(max_length=models.Application.tag.property.columns[0].type.length)
    capture_kit: Optional[str]
    collection_date: Optional[str]
    comment: Optional[constr(max_length=models.Sample.comment.property.columns[0].type.length)]
    concentration: Optional[float]
    concentration_sample: Optional[float]
    container: Optional[ContainerEnum]
    container_name: Optional[str]
    control: Optional[str]
    customer: Optional[
        constr(max_length=models.Customer.internal_id.property.columns[0].type.length)
    ]
    custom_index: Optional[str]
    data_analysis: Pipeline
    data_delivery: DataDelivery
    elution_buffer: Optional[str]
    extraction_method: Optional[str]
    family_name: Optional[
        constr(
            regex=NAME_PATTERN,
            min_length=2,
            max_length=models.Family.name.property.columns[0].type.length,
        )
    ]
    father: Optional[
        constr(regex=NAME_PATTERN, max_length=models.Sample.name.property.columns[0].type.length)
    ]
    formalin_fixation_time: Optional[int]
    index: Optional[str]
    index_number: Optional[str]
    index_sequence: Optional[str]
    internal_id: Optional[
        constr(max_length=models.Sample.internal_id.property.columns[0].type.length)
    ]
    lab_code: Optional[str]
    mother: Optional[
        constr(regex=NAME_PATTERN, max_length=models.Sample.name.property.columns[0].type.length)
    ]
    name: constr(
        regex=NAME_PATTERN,
        min_length=2,
        max_length=models.Sample.name.property.columns[0].type.length,
    )
    organism: Optional[str]
    organism_other: Optional[str]
    original_lab: Optional[str]
    original_lab_address: Optional[str]
    phenotype_groups: Optional[List[str]]
    phenotype_terms: Optional[List[str]]
    pool: Optional[constr(max_length=models.Pool.name.property.columns[0].type.length)]
    post_formalin_fixation_time: Optional[int]
    pre_processing_method: Optional[str]
    priority: PriorityEnum = PriorityEnum.standard
    quantity: Optional[int]
    reagent_label: Optional[str]
    reference_genome: Optional[
        constr(max_length=models.Sample.reference_genome.property.columns[0].type.length)
    ]
    region: Optional[str]
    region_code: Optional[str]
    require_qcok: bool = False
    rml_plate_name: Optional[str]
    selection_criteria: Optional[str]
    sex: SexEnum = SexEnum.unknown
    source: Optional[str]
    status: StatusEnum = StatusEnum.unknown
    subject_id: Optional[
        constr(
            regex=NAME_PATTERN, max_length=models.Sample.subject_id.property.columns[0].type.length
        )
    ]
    tissue_block_size: Optional[str]
    tumour: bool = False
    tumour_purity: Optional[int]
    verified_organism: Optional[bool]
    volume: Optional[str]
    well_position: Optional[str]
    well_position_rml: Optional[str]

    @validator("priority", pre=True)
    def snake_case(cls, value: str):
        return value.lower().replace(" ", "_")



"""Tests for rsync API"""
import logging
from typing import List

import pytest
from pathlib import Path

from cgmodels.cg.constants import Pipeline
from cg.exc import CgError
from cg.meta.rsync import RsyncAPI
from cg.models.cg_config import CGConfig
from cg.store import models, Store


def test_get_source_and_destination_paths(
    mutant_case: models.Family, rsync_api: RsyncAPI, ticket: str, mocker
):
    """Test generating the source path before rsync"""

    # GIVEN a valid Sars-cov-2 case
    case = mutant_case

    # GIVEN file exists
    mocker.patch.object(RsyncAPI, "get_all_cases_from_ticket")
    RsyncAPI.get_all_cases_from_ticket.return_value = [case]

    # WHEN the source path is created
    source_and_destination_paths = rsync_api.get_source_and_destination_paths(ticket=ticket)

    # THEN the source path ends with a customer id, followed by "inbox" and a ticket id
    assert (
        source_and_destination_paths["delivery_source_path"]
        .as_posix()
        .endswith(f"/cust000/inbox/{ticket}")
    )
    # THEN the destination path is in the format server.name.se:/path/cust_id/path/ticket/
    assert (
        source_and_destination_paths["rsync_destination_path"].as_posix()
        == "server.name.se:/some/cust000/inbox"
    )


def test_get_source_path_no_case(rsync_api: RsyncAPI, ticket: str, mocker, helpers, caplog):
    """Test generating the source path before rsync when there is no case"""
    caplog.set_level(logging.WARNING)

    # GIVEN file exists
    mocker.patch.object(RsyncAPI, "get_all_cases_from_ticket")
    RsyncAPI.get_all_cases_from_ticket.return_value = None

    with pytest.raises(CgError):
        # WHEN the source path is collected
        rsync_api.get_source_and_destination_paths(ticket=ticket)

        # THEN the source path ends with a customer id, followed by "inbox" and a ticket id
        assert "Could not find any cases for ticket_id" in caplog.text


def test_set_log_dir(rsync_api: RsyncAPI, ticket: str, caplog):
    """Test function to set log dir for path"""

    caplog.set_level(logging.INFO)

    # GIVEN an RsyncAPI, with its base path as its log dir
    base_path: Path = rsync_api.log_dir

    # WHEN setting the log directory
    rsync_api.set_log_dir(folder_prefix=ticket)

    # THEN the log dir should set to a new path, different from the base path
    assert base_path.as_posix() != rsync_api.log_dir.as_posix()
    assert "Setting log dir to:" in caplog.text


def test_make_log_dir(rsync_api: RsyncAPI, ticket: str, caplog):
    """Test generating the directory for logging"""
    caplog.set_level(logging.INFO)

    # WHEN the log directory is created
    rsync_api.set_log_dir(folder_prefix=ticket)
    rsync_api.create_log_dir(dry_run=True)

    # THEN the path is not created since it is a dry run
    assert "Would have created path" in caplog.text

    # THEN the created path is
    assert str(rsync_api.log_dir).startswith(f"/another/path/{ticket}")


def test_run_rsync_on_slurm(
    microsalt_case: models.Family, rsync_api: RsyncAPI, ticket: str, caplog, mocker, helpers
):
    """Test for running rsync on slurm"""
    caplog.set_level(logging.INFO)

    # GIVEN a valid microsalt case
    case: models.Family = microsalt_case

    # GIVEN paths needed to run rsync
    mocker.patch.object(RsyncAPI, "get_source_and_destination_paths")
    RsyncAPI.get_source_and_destination_paths.return_value = {
        "delivery_source_path": Path("/path/to/source"),
        "rsync_destination_path": Path("/path/to/destination"),
    }

    mocker.patch.object(RsyncAPI, "get_all_cases_from_ticket")
    RsyncAPI.get_all_cases_from_ticket.return_value = [case]

    # WHEN the destination path is created
    sbatch_number: int = rsync_api.run_rsync_on_slurm(ticket=ticket, dry_run=True)

    # THEN check that SARS-COV-2 analysis is not delivered
    assert "Delivering report for SARS-COV-2 analysis" not in caplog.text

    # THEN check that an integer was returned as sbatch number
    assert isinstance(sbatch_number, int)


def test_get_folders_to_deliver(
    analysis_family: dict, analysis_store_trio, rsync_api: RsyncAPI, case_id: str
):
    """Tests the ability for the rsync api to get case and sample names"""
    # GIVEN a case

    # WHEN the function gets the folders
    folder_list: List[str] = rsync_api.get_folders_to_deliver(
        case_id=case_id, sample_files_present=True, case_files_present=True
    )

    # THEN it the list should contain the case name and all the samples
    assert folder_list == [
        analysis_family["samples"][0]["name"],
        analysis_family["samples"][1]["name"],
        analysis_family["samples"][2]["name"],
        analysis_family["name"],
    ]


def test_concatenate_rsync_commands(
    analysis_family: dict, analysis_store_trio, project_dir, customer_id, ticket: str
):
    """Tests the function to concatenate rsync commands for transferring multiple files"""
    # GIVEN a list with a case and a sample name
    folder_list: List[str] = [analysis_family["name"], analysis_family["samples"][0]["name"]]
    source_and_destination_paths = {
        "delivery_source_path": project_dir / customer_id / ticket,
        "rsync_destination_path": project_dir / customer_id,
    }
    # WHEN then commands are generated
    commands: str = RsyncAPI.concatenate_rsync_commands(
        folder_list=folder_list,
        source_and_destination_paths=source_and_destination_paths,
        ticket=ticket,
    )
    # THEN the correct folder should be added to the source path
    assert (
        " ".join(
            [
                str(source_and_destination_paths["delivery_source_path"] / analysis_family["name"]),
                str(source_and_destination_paths["delivery_source_path"]),
            ]
        )
        in commands
    )
    assert (
        " ".join(
            [
                str(
                    source_and_destination_paths["delivery_source_path"]
                    / analysis_family["samples"][0]["name"]
                ),
                str(source_and_destination_paths["delivery_source_path"]),
            ]
        )
        in commands
    )


def test_slurm_rsync_single_case(
    microsalt_case: models.Family, rsync_api: RsyncAPI, caplog, mocker, helpers, ticket: str
):
    """Test for running rsync on a single case on slurm"""
    caplog.set_level(logging.INFO)

    # GIVEN a valid microsalt case
    case: models.Family = microsalt_case

    # GIVEN paths needed to run rsync
    mocker.patch.object(RsyncAPI, "get_source_and_destination_paths")
    RsyncAPI.get_source_and_destination_paths.return_value = {
        "delivery_source_path": Path("/path/to/source"),
        "rsync_destination_path": Path("/path/to/destination"),
    }

    mocker.patch.object(Store, "get_latest_ticket_from_case")
    Store.get_latest_ticket_from_case.return_value = ticket

    # WHEN the destination path is created
    sbatch_number: int = rsync_api.slurm_rsync_single_case(
        case_id=case.internal_id, case_files_present=True, dry_run=True
    )

    # THEN check that an integer was returned as sbatch number
    assert isinstance(sbatch_number, int)



from pathlib import Path

import pytest


@pytest.fixture(name="spring_file_path")
def fixture_spring_file_path() -> Path:
    """Return spring file path"""
    return Path("/path/to/spring_file.spring")


@pytest.fixture(name="backup_file_path")
def fixture_backup_file_path() -> str:
    """Return path to a file used in the backup process"""
    return "/path/to/backup_file.extension"


@pytest.fixture(name="archived_flow_cell")
def fixture_archived_flow_cell() -> Path:
    """Path of archived flow cell"""
    return Path("/path/to/archived/flow_cell.tar.gz.gpg")


@pytest.fixture(name="archived_key")
def fixture_archived_key() -> Path:
    """Path of archived key"""
    return Path("/path/to/archived/encryption_key.key.gpg")


@pytest.fixture(name="pdc_query")
def fixture_pdc_query() -> list:
    """Result of a PDC query"""
    return [
        "IBM Tivoli Storage Manager",
        "Command Line Backup-Archive Client Interface",
        "  Client Version 7, Release 1, Level 4.0 ",
        "  Client date/time: 05/15/2022 11:43:55",
        "(c) Copyright by IBM Corporation and other(s) 1990, 2015. All Rights Reserved.",
        "",
        "Node Name: HASTA.SCILIFELAB.SE_CLINICAL",
        "Session established with server BLACKHOLE: Linux/x86_64",
        "  Server Version 8, Release 1, Level 9.300",
        "  Server date/time: 05/15/2022 11:43:55  Last access: 05/15/2022 10:57:20",
        "",
        "Accessing as node: SLLCLINICAL",
        "             Size  Archive Date - Time    File - Expires on - Description",
        "             ----  -------------------    -------------------------------",
        "           607  B  03/28/2022 11:59:54    "
        "/path/to/archived/encryption_key.key.gpg Never "
        "Archive Date: 03/28/2022",
        "43,868,543,141  B  03/28/2022 11:53:04    "
        "/path/to/archived/flow_cell.tar.gz.gpg Never "
        "Archive Date: 03/28/2022",
    ]



"""MIP-DNA upload API"""

import logging

import click

from cg.cli.generate.report.base import delivery_report
from cg.cli.upload.scout import scout
from cg.cli.upload.observations import observations
from cg.cli.upload.genotype import genotypes
from cg.cli.upload.validate import validate
from cg.cli.upload.coverage import coverage
from cg.cli.upload.clinical_delivery import clinical_delivery
from cg.constants import DataDelivery, REPORT_SUPPORTED_DATA_DELIVERY
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.meta.upload.upload_api import UploadAPI
from cg.store import models
from cg.apps.tb import TrailblazerAPI


LOG = logging.getLogger(__name__)


class MipDNAUploadAPI(UploadAPI):
    """MIP-DNA upload API"""

    def __init__(self, config: CGConfig):
        self.analysis_api: MipDNAAnalysisAPI = MipDNAAnalysisAPI(config)
        super().__init__(config=config, analysis_api=self.analysis_api)

    def upload(self, ctx: click.Context, case_obj: models.Family, restart: bool) -> None:
        """Uploads MIP-DNA analysis data and files"""

        analysis_obj: models.Analysis = case_obj.analyses[0]
        self.update_upload_started_at(analysis_obj)

        # Main upload
        ctx.invoke(clinical_delivery, case_id=case_obj.internal_id)
        ctx.invoke(coverage, family_id=case_obj.internal_id, re_upload=restart)
        ctx.invoke(validate, family_id=case_obj.internal_id)
        ctx.invoke(genotypes, family_id=case_obj.internal_id, re_upload=restart)
        ctx.invoke(observations, case_id=case_obj.internal_id)

        # Delivery report generation
        if case_obj.data_delivery in REPORT_SUPPORTED_DATA_DELIVERY:
            ctx.invoke(delivery_report, case_id=case_obj.internal_id)

        # Scout specific upload
        if DataDelivery.SCOUT in case_obj.data_delivery:
            ctx.invoke(scout, case_id=case_obj.internal_id, re_upload=restart)

        self.update_uploaded_at(analysis_obj)



from pathlib import Path
from typing import Optional
from pydantic import BaseModel, validator
from datetime import date, datetime

from cg.meta.upload.gisaid.constants import AUTHORS


class FastaFile(BaseModel):
    header: str
    sequence: str


class GisaidAccession(BaseModel):
    log_message: str
    accession_nr: Optional[str]
    sample_id: Optional[str]

    @validator("accession_nr", always=True)
    def parse_accession(cls, v, values):
        return values["log_message"].split(";")[-1]

    @validator("sample_id", always=True)
    def parse_sample_id(cls, v, values):
        return values["log_message"].split("/")[2].split("_")[2]


class UploadFiles(BaseModel):
    csv_file: Path
    fasta_file: Path
    log_file: Path


class GisaidSample(BaseModel):
    case_id: str
    cg_lims_id: str
    submitter: str
    region: str
    region_code: str
    fn: str
    covv_collection_date: str
    covv_subm_sample_id: str
    covv_virus_name: Optional[str]
    covv_orig_lab: Optional[str]
    covv_type: Optional[str] = "betacoronavirus"
    covv_passage: Optional[str] = "Original"
    covv_location: Optional[str]
    covv_host: Optional[str] = "Human"
    covv_gender: Optional[str] = "unknown"
    covv_patient_age: Optional[str] = "unknown"
    covv_patient_status: Optional[str] = "unknown"
    covv_seq_technology: Optional[str] = "Illumina NovaSeq"
    covv_orig_lab_addr: Optional[str]
    covv_subm_lab: Optional[str] = "Karolinska University Hospital"
    covv_subm_lab_addr: Optional[str] = "171 76 Stockholm, Sweden"
    covv_authors: Optional[str] = " ,".join(AUTHORS)

    @validator("covv_location", always=True)
    def parse_location(cls, v, values):
        region: str = values.get("region")
        return f"Europe/Sweden/{region}"

    @validator("covv_subm_sample_id", always=True)
    def parse_subm_sample_id(cls, v, values):
        region_code = values.get("region_code")
        return f"{region_code}_SE100_{v}"

    @validator("covv_virus_name", always=True)
    def parse_virus_name(cls, v, values):
        sample_name = values.get("covv_subm_sample_id")
        date = values.get("covv_collection_date")
        datetime_date = datetime.strptime(date, "%Y-%m-%d")
        return f"hCoV-19/Sweden/{sample_name}/{datetime_date.year}"



"""This script tests the cli methods to get flowcells in status-db"""
from datetime import datetime

from cg.cli.get import get
from cg.models.cg_config import CGConfig
from cg.store import Store
from click.testing import CliRunner
from tests.store_helpers import StoreHelpers


def test_get_flowcell_bad_flowcell(cli_runner: CliRunner, base_context: CGConfig):
    """Test to get a flow cell using a non-existing flow cell"""
    # GIVEN an empty database

    # WHEN getting a flow cell
    name = "dummy_name"
    result = cli_runner.invoke(get, ["flowcell", name], obj=base_context)

    # THEN it should complain in missing flow cell instead of getting a flow cell
    assert result.exit_code == 1


def test_get_flowcell_required(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test to get a flow cell using only the required arguments."""
    # GIVEN a database with a flow cell
    flow_cell = helpers.add_flowcell(disk_store)
    flow_cell_name = flow_cell.name
    assert disk_store.Flowcell.query.count() == 1

    # WHEN getting a flow cell

    result = cli_runner.invoke(get, ["flowcell", flow_cell_name], obj=base_context)

    # THEN it should have been get
    assert result.exit_code == 0


def test_get_flowcell_output(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output has the data of the flow cell."""
    # GIVEN a database with a flow cell with data
    flow_cell = helpers.add_flowcell(disk_store)
    flow_cell_name = flow_cell.name
    sequencer_type = disk_store.Flowcell.query.first().sequencer_type
    sequencer_name = disk_store.Flowcell.query.first().sequencer_name
    sequenced_at_date = str(disk_store.Flowcell.query.first().sequenced_at.date())
    status = disk_store.Flowcell.query.first().status

    # WHEN getting a flowcell
    result = cli_runner.invoke(get, ["flowcell", flow_cell_name], obj=base_context)

    # THEN it should have been get
    assert result.exit_code == 0
    assert flow_cell_name in result.output
    assert sequencer_type in result.output
    assert sequencer_name in result.output
    assert sequenced_at_date in result.output
    assert status in result.output


def test_get_flowcell_archived_at_none(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output has the data of the flow cell."""
    # GIVEN a database with a flowcell with data
    flowcell = helpers.add_flowcell(disk_store, archived_at=None)
    archived_at = "No"

    # WHEN getting a flowcell
    result = cli_runner.invoke(get, ["flowcell", flowcell.name], obj=base_context)

    # THEN it should have been get
    assert result.exit_code == 0
    assert archived_at in result.output


def test_get_flowcell_archived_at_date(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output has the data of the flow cell."""
    # GIVEN a database with a flowcell with data
    archived_at = datetime.now()
    flowcell = helpers.add_flowcell(disk_store, archived_at=archived_at)
    archived_at_date = str(archived_at.date())

    # WHEN getting a flowcell
    result = cli_runner.invoke(get, ["flowcell", flowcell.name], obj=base_context)

    # THEN it should have been get
    assert result.exit_code == 0
    assert archived_at_date in result.output


def test_get_flowcell_samples_without_samples(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers, caplog
):
    """Test that the output has the data of the flow cell."""
    # GIVEN a database with a flowcell without related samples
    flowcell = helpers.add_flowcell(disk_store)
    assert not disk_store.Flowcell.query.first().samples

    # WHEN getting a flowcell with the --samples flag
    result = cli_runner.invoke(get, ["flowcell", flowcell.name, "--samples"], obj=base_context)

    # THEN a message about no samples should have been displayed
    assert result.exit_code == 0
    assert "no samples found on flowcell" in caplog.text


def test_get_flowcell_samples(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output has the data of the flow cell."""
    # GIVEN a database with a flowcell with related samples
    samples = helpers.add_samples(disk_store)
    flowcell = helpers.add_flowcell(disk_store, samples=samples)
    assert disk_store.Flowcell.query.first().samples

    # WHEN getting a flowcell with the --samples flag
    result = cli_runner.invoke(get, ["flowcell", flowcell.name, "--samples"], obj=base_context)

    # THEN all related samples should be listed in the output
    assert result.exit_code == 0
    for sample in disk_store.Flowcell.query.first().samples:
        assert sample.internal_id in result.output


def test_get_flowcell_no_samples_without_samples(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output has the data of the flow cell."""
    # GIVEN a database with a flowcell without related samples
    flowcell = helpers.add_flowcell(disk_store)
    assert not disk_store.Flowcell.query.first().samples

    # WHEN getting a flowcell with the --no-samples flag
    result = cli_runner.invoke(get, ["flowcell", flowcell.name, "--no-samples"], obj=base_context)

    # THEN there are no samples to display but everything is OK
    assert result.exit_code == 0


def test_get_flowcell_no_samples_with_samples(
    cli_runner: CliRunner, base_context: CGConfig, disk_store: Store, helpers: StoreHelpers
):
    """Test that the output has the data of the flow cell."""
    # GIVEN a database with a flowcell with related samples
    samples = helpers.add_samples(disk_store)
    flowcell = helpers.add_flowcell(disk_store, samples=samples)
    assert disk_store.Flowcell.query.first().samples

    # WHEN getting a flowcell with the --no-samples flag
    db_uri = disk_store.uri

    result = cli_runner.invoke(get, ["flowcell", flowcell.name, "--no-samples"], obj=base_context)

    # THEN no related samples should be listed in the output
    assert result.exit_code == 0
    for sample in disk_store.Flowcell.query.first().samples:
        assert sample.internal_id not in result.output



from datetime import datetime

from cg.constants import FlowCellStatus
from cg.constants.sequencing import Sequencers
from cg.store import Store
from cg.store.models import Flowcell
from tests.store_helpers import StoreHelpers


def test_add_flowcell(
    base_store: Store,
    flow_cell_id: str,
    helpers: StoreHelpers,
    sequencer_name: str,
    timestamp_now: datetime,
):
    """Test adding a flow cell to the database."""

    # GIVEN a database with no flow cell

    # WHEN adding flow cell
    flow_cell: Flowcell = base_store.add_flow_cell(
        flow_cell_id=flow_cell_id,
        sequencer_name=sequencer_name,
        sequencer_type=Sequencers.NOVASEQ,
        date=timestamp_now,
        flow_cell_status=FlowCellStatus.ONDISK,
    )

    # THEN flow cell should be returned
    assert flow_cell

    # THEN the flow cell status should be "ondisk"
    assert flow_cell.status == FlowCellStatus.ONDISK


def test_add_flowcell_status(
    base_store: Store,
    flow_cell_id: str,
    helpers: StoreHelpers,
    sequencer_name: str,
    timestamp_now: datetime,
):
    """Test adding a flow cell with a status to the database."""

    # GIVEN a database with no flow cell

    # WHEN adding flow cell
    flow_cell: Flowcell = base_store.add_flow_cell(
        flow_cell_id=flow_cell_id,
        sequencer_name=sequencer_name,
        sequencer_type=Sequencers.NOVASEQ,
        date=timestamp_now,
        flow_cell_status=FlowCellStatus.PROCESSING,
    )

    # THEN the flow cell status should be "processing"
    assert flow_cell.status == FlowCellStatus.PROCESSING



"""Observations API."""

import logging
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Union

from housekeeper.store.models import Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.loqus import LoqusdbAPI
from cg.constants.observations import LoqusdbInstance
from cg.models.cg_config import CGConfig, CommonAppConfig
from cg.models.observations.input_files import (
    MipDNAObservationsInputFiles,
    BalsamicObservationsInputFiles,
)
from cg.store import Store, models

LOG = logging.getLogger(__name__)


class ObservationsAPI:
    """API to manage Loqusdb observations."""

    def __init__(self, config: CGConfig):
        self.store: Store = config.status_db
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.loqusdb_config: CommonAppConfig = config.loqusdb
        self.loqusdb_wes_config: CommonAppConfig = config.loqusdb_wes
        self.loqusdb_somatic_config: CommonAppConfig = config.loqusdb_somatic
        self.loqusdb_tumor_config: CommonAppConfig = config.loqusdb_tumor

    def upload(self, case: models.Family) -> None:
        """Upload observations to Loqusdb."""
        input_files: Union[
            MipDNAObservationsInputFiles, BalsamicObservationsInputFiles
        ] = self.get_observations_input_files(case)
        self.load_observations(case=case, input_files=input_files)

    def get_observations_input_files(
        self, case: models.Family
    ) -> Union[MipDNAObservationsInputFiles, BalsamicObservationsInputFiles]:
        """Fetch input files from a case to upload to Loqusdb."""
        analysis: models.Analysis = case.analyses[0]
        analysis_date: datetime = analysis.started_at or analysis.completed_at
        hk_version: Version = self.housekeeper_api.version(
            analysis.family.internal_id, analysis_date
        )
        return self.extract_observations_files_from_hk(hk_version)

    def get_loqusdb_api(self, loqusdb_instance: LoqusdbInstance) -> LoqusdbAPI:
        """Returns a Loqusdb API for the given Loqusdb instance."""
        loqusdb_apis = {
            LoqusdbInstance.WGS: LoqusdbAPI(
                binary_path=self.loqusdb_config.binary_path,
                config_path=self.loqusdb_config.config_path,
            ),
            LoqusdbInstance.WES: LoqusdbAPI(
                binary_path=self.loqusdb_wes_config.binary_path,
                config_path=self.loqusdb_wes_config.config_path,
            ),
            LoqusdbInstance.SOMATIC: LoqusdbAPI(
                binary_path=self.loqusdb_somatic_config.binary_path,
                config_path=self.loqusdb_somatic_config.config_path,
            ),
            LoqusdbInstance.TUMOR: LoqusdbAPI(
                binary_path=self.loqusdb_tumor_config.binary_path,
                config_path=self.loqusdb_tumor_config.config_path,
            ),
        }
        return loqusdb_apis[loqusdb_instance]

    @staticmethod
    def is_duplicate(
        case: models.Family,
        loqusdb_api: LoqusdbAPI,
        profile_vcf_path: Path,
        profile_threshold: float,
    ) -> bool:
        """Check if a case has already been uploaded to Loqusdb."""
        loqusdb_case: dict = loqusdb_api.get_case(case_id=case.internal_id)
        duplicate = loqusdb_api.get_duplicate(
            profile_vcf_path=profile_vcf_path, profile_threshold=profile_threshold
        )
        return bool(loqusdb_case or duplicate or case.loqusdb_uploaded_samples)

    def update_statusdb_loqusdb_id(
        self, samples: List[models.Family], loqusdb_id: Optional[str]
    ) -> None:
        """Update Loqusdb ID field in StatusDB for each of the provided samples."""
        for sample in samples:
            sample.loqusdb_id = loqusdb_id
        self.store.commit()

    def load_observations(
        self,
        case: models.Family,
        input_files: Union[MipDNAObservationsInputFiles, BalsamicObservationsInputFiles],
    ) -> None:
        """Load observation counts to Loqusdb."""
        raise NotImplementedError

    def extract_observations_files_from_hk(
        self, hk_version: Version
    ) -> Union[MipDNAObservationsInputFiles, BalsamicObservationsInputFiles]:
        """Extract observations files given a housekeeper version."""
        raise NotImplementedError

    def delete_case(self, case: models.Family) -> None:
        """Delete case observations from Loqusdb."""
        raise NotImplementedError



""" Test the osticket app """

import logging

import pytest

from cg.apps.osticket import OsTicket
from cg.exc import TicketCreationError


def test_osticket_respone_500(monkeypatch, caplog, response):
    """Test that we log properly when we get a non successful response"""

    # GIVEN a ticket server always gives a failure in response
    osticket_api = OsTicket()
    monkeypatch.setattr("requests.post", lambda *args, **kwargs: response)

    # WHEN we call open_ticket with ok ticket data
    caplog.set_level(logging.ERROR)
    with pytest.raises(TicketCreationError):
        osticket_api.open_ticket(
            name="dummy_name",
            email="dummy_email",
            subject="dummy_subject",
            message="dummy_message",
            attachment={},
        )

    # THEN the response text and reason was logged and a ticket creation error raised
    assert response.reason in caplog.text
    assert response.text in caplog.text
